#!/usr/bin/env python

import math
import sys
import random

pi=math.pi
e=math.e
j=complex(0,1)

def fft(f,inv):
    n=len(f)
    if n==1:
        return f

    for p in 2,3,5:
        if n%p==0:
            break
    else:
        raise Exception('%s not factorable ' % n)

    m = n/p
    Fout=[]
    for q in range(p): # 0,1
        fp = f[q::p]  # every p'th time sample
        Fp = fft( fp ,inv)
        Fout.extend( Fp )

    for u in range(m):
        scratch = Fout[u::m] # u to end in strides of m
        for q1 in range(p):
            k = q1*m + u  # indices to Fout above that became scratch
            Fout[ k ] = scratch[0] # cuz e**0==1 in loop below
            for q in range(1,p):
                if inv:
                    t = e ** ( j*2*pi*k*q/n )
                else:                    
                    t = e ** ( -j*2*pi*k*q/n )
                Fout[ k ] += scratch[q] * t

    return Fout

def rifft(F):
    N = len(F) - 1
    Z = [0] * (N)
    for k in range(N):
        Fek = ( F[k] + F[-k-1].conjugate() )
        Fok = ( F[k] - F[-k-1].conjugate() ) * e ** (j*pi*k/N)
        Z[k] = Fek + j*Fok

    fp = fft(Z , 1)

    f = []
    for c in fp:
        f.append(c.real)
        f.append(c.imag)
    return f

def real_fft( f,inv ):
    if inv:
        return rifft(f)

    N = len(f) / 2

    res = f[::2]
    ims = f[1::2]

    fp = [ complex(r,i) for r,i in zip(res,ims) ]
    print 'fft input ', fp
    Fp = fft( fp ,0 )
    print 'fft output ', Fp

    F = [ complex(0,0) ] * ( N+1 )
    
    F[0] = complex( Fp[0].real + Fp[0].imag , 0 ) 

    for k in range(1,N/2+1):
        tw = e ** ( -j*pi*(.5+float(k)/N ) )
        
        F1k = Fp[k] + Fp[N-k].conjugate()
        F2k = Fp[k] - Fp[N-k].conjugate()
        F2k *= tw
        F[k] = ( F1k + F2k ) * .5
        F[N-k] = ( F1k - F2k ).conjugate() * .5
        #F[N-k] = ( F1kp + e ** ( -j*pi*(.5+float(N-k)/N ) ) * F2kp ) * .5
        #F[N-k] = ( F1k.conjugate() - tw.conjugate() * F2k.conjugate() ) * .5

    F[N] = complex( Fp[0].real - Fp[0].imag , 0 ) 
    return F

def main():
    #fft_func = fft
    fft_func = real_fft

    tvec = [0.309655,0.815653,0.768570,0.591841,0.404767,0.637617,0.007803,0.012665]
    Ftvec = [ complex(r,i) for r,i in zip(
                [3.548571,-0.378761,-0.061950,0.188537,-0.566981,0.188537,-0.061950,-0.378761],
                [0.000000,-1.296198,-0.848764,0.225337,0.000000,-0.225337,0.848764,1.296198] ) ]

    F = fft_func( tvec,0 )

    nerrs= 0
    for i in range(len(Ftvec)/2 + 1):
        if abs( F[i] - Ftvec[i] )> 1e-5:
            print 'F[%d]: %s != %s' % (i,F[i],Ftvec[i])
            nerrs += 1

    print '%d errors in forward fft' % nerrs
    if nerrs:
        return

    trec = fft_func( F , 1 )

    for i in range(len(trec) ):
        trec[i] /= len(trec)

    for i in range(len(tvec) ):
        if abs( trec[i] - tvec[i] )> 1e-5:
            print 't[%d]: %s != %s' % (i,tvec[i],trec[i])
            nerrs += 1

    print '%d errors in reverse fft' % nerrs


def make_random(dims=[1]):
    import Numeric 
    res = []
    for i in range(dims[0]):
        if len(dims)==1:
            r=random.uniform(-1,1)
            i=random.uniform(-1,1)
            res.append( complex(r,i) )
        else:
            res.append( make_random( dims[1:] ) )
    return Numeric.array(res)

def flatten(x):
    import Numeric
    ntotal = Numeric.product(Numeric.shape(x))
    return Numeric.reshape(x,(ntotal,))

def randmat( ndims ):
    dims=[]
    for i in range( ndims ):
        curdim = int( random.uniform(2,4) )
        dims.append( curdim )
    return make_random(dims )

def test_fftnd(ndims=3):
    import FFT
    import Numeric

    x=randmat( ndims )
    print 'dimensions=%s' % str( Numeric.shape(x) )
    #print 'x=%s' %str(x)
    xver = FFT.fftnd(x)
    x2=myfftnd(x)
    err = xver - x2
    errf = flatten(err)
    xverf = flatten(xver)
    errpow = Numeric.vdot(errf,errf)+1e-10
    sigpow = Numeric.vdot(xverf,xverf)+1e-10
    snr = 10*math.log10(abs(sigpow/errpow) )
    if snr<80:
        print xver
        print x2
    print 'SNR=%sdB' % str( snr )
 
def myfftnd(x):
    import Numeric
    xf = flatten(x)
    Xf = fftndwork( xf , Numeric.shape(x) )
    return Numeric.reshape(Xf,Numeric.shape(x) )

def fftndwork(x,dims):
    import Numeric
    dimprod=Numeric.product( dims )

    for k in range( len(dims) ):
        cur_dim=dims[ k ]
        stride=dimprod/cur_dim
        next_x = [complex(0,0)]*len(x)
        for i in range(stride):
            next_x[i*cur_dim:(i+1)*cur_dim] = fft(x[i:(i+cur_dim)*stride:stride],0)
        x = next_x
    return x

if __name__ == "__main__":
    try:
        nd = int(sys.argv[1])
    except:
        nd=None
    if nd:    
        test_fftnd( nd )
    else:    
        sys.exit(0)



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from itertools import izip
import numpy
import essentia
from essentia import EssentiaError
import numpy.core.arrayprint

# make it so we don't see summarized output, we want the full stuff
numpy.core.arrayprint.set_printoptions(precision=10,
                                       linewidth=80,
                                       threshold=1e6,
                                       suppress=True)


def percentile(values, p):
    values.sort()
    idx = int(p/100.0 * len(values))
    return values[idx]


# a simple pool of numbers with save to audioclas format and save to textfile
class Pool:
    GlobalScope = -1   # this value doesn't really matter, we just need the identifier
    CurrentScope = -2  # this value doesn't really matter, we just need the identifier

    def __init__(self):
        self.descriptors = {}
        self.statsBlackList = []

        self.__currentNamespace = 'AnonymousNamespace'
        self.__globalScope = [0.0, 0.0]
        self.__currentScope = [0.0, 0.0]

    def setCurrentNamespace(self, namespace):
        self.__currentNamespace = namespace
        if namespace not in self.descriptors:
            self.descriptors[namespace] = {}

    def ns(self):
        return self.__currentNamespace

    def setGlobalScope(self, scope):
        self.__globalScope = scope

    def setCurrentScope(self, scope):
        self.__currentScope = scope

    def add(self, name, value, scope=CurrentScope):
        if isinstance(value, tuple):
            raise EssentiaError('error when adding %s: you can\'t add tuples to the pool' % name)

        if scope == self.GlobalScope:
            self.statsBlackList.append((self.__currentNamespace, name))
            scope = self.__globalScope

        elif scope == self.CurrentScope:
            scope = self.__currentScope

        try:
            self.descriptors[self.__currentNamespace][name]['values'].append(value)
            self.descriptors[self.__currentNamespace][name]['scopes'].append(scope)
        except KeyError:
            self.descriptors[self.__currentNamespace][name] = {'values': [value], 'scopes': [scope]}

    def mean_scope(self, scopeFrom, scopeTo):
        descriptors_mean = {}

        for key in self.descriptors.keys():
            descriptor = self.descriptors[self.__currentNamespace][key]
            values_in_scope = []

            # Global descriptor (should also check that scope spans the entire file)
            if len(descriptor['values']) == 1:
                descriptors_mean[key] = descriptor['values'][0]
                continue

            for scope, value in zip(descriptor['scopes'], descriptor['values']):
                if scope[0] >= scopeFrom and scope[1] <= scopeTo:
                    values_in_scope.append(value)

            if len(values_in_scope) > 0:
                try:
                    descriptors_mean[key] = essentia.array(numpy.mean(values_in_scope, axis=0))
                except TypeError:  # values are not numeric
                    descriptors_mean[key] = values_in_scope[0]

        return descriptors_mean

    def var_scope(self, scopeFrom, scopeTo):
        descriptors_var = {}

        for key in self.descriptors.keys():
            descriptor = self.descriptors[self.__currentNamespace][key]
            values_in_scope = []

            # Global descriptor (should also check that scope spans the entire file)
            if len(descriptor['values']) == 1:
                descriptors_var[key] = 0.0
                continue

            for scope, value in zip(descriptor['scopes'], descriptor['values']):
                if scope[0] >= scopeFrom and scope[1] <= scopeTo:
                    values_in_scope.append(value)

            if len(values_in_scope) > 0:
                try:
                    descriptors_var[key] = essentia.array(numpy.var(values_in_scope, axis=0))
                except TypeError:  # values are not numeric
                    descriptors_var[key] = 0.0

        return descriptors_var

    def aggregate_descriptors(self, descriptors_stats={}):
        aggregated = {}
        for namespace in self.descriptors:
            aggregated[namespace] = {}
            descs = self.descriptors[namespace].keys()
            descs.sort()

            stats_default = ['mean', 'var', 'min', 'max']

            for desc in descs:
                values = self.descriptors[namespace][desc]['values']

                if (namespace, desc) in self.statsBlackList:
                    # make sure there is only one value
                    if len(values) != 1:
                        raise EssentiaError('You declared %s as a global descriptors, but there are more than 1 value: %s' % (desc, values))

                    value = values[0]
                    try:
                        # if value is numeric
                        aggregated[namespace][desc] = {'value': essentia.array(value)}
                    except:
                        # if value is not numeric
                        aggregated[namespace][desc] = {'value': value}

                    continue

                aggregated[namespace][desc] = {}
                aggrDesc = aggregated[namespace][desc]

                stats = list(stats_default)
                if not isinstance(values[0], numpy.ndarray):
                    #stats += [ 'percentile_5', 'percentile_95' ]
                    stats += ['dmean', 'dmean2', 'dvar', 'dvar2']

                if namespace in descriptors_stats and desc in descriptors_stats[namespace]:
                    stats = descriptors_stats[namespace][desc]

                try:

                    if 'mean' in stats:
                        aggrDesc['mean'] = essentia.array(numpy.mean(values, axis=0))

                    if 'var' in stats:
                        aggrDesc['var'] = essentia.array(numpy.var(values, axis=0))

                    if 'min' in stats:
                        aggrDesc['min'] = essentia.array(numpy.min(values, axis=0))

                    if 'max' in stats:
                        aggrDesc['max'] = essentia.array(numpy.max(values, axis=0))

                    derived = None
                    derived2 = None

                    if 'dmean' in stats:
                        if not derived:
                            derived = [a - b for a, b in izip(values[1:], values[:-1])]
                        aggrDesc['dmean'] = essentia.array(numpy.mean(numpy.abs(derived), axis=0))

                    if 'dvar' in stats:
                        if not derived:
                            derived = [a - b for a, b in izip(values[1:], values[:-1])]
                        aggrDesc['dvar'] = essentia.array(numpy.var(derived, axis=0))

                    if 'dmean2' in stats:
                        if not derived:
                            derived = [a - b for a, b in izip(values[1:], values[:-1])]
                        if not derived2:
                            derived2 = [a - b for a, b in izip(derived[1:], derived[:-1])]
                        if derived2:
                            aggrDesc['dmean2'] = essentia.array(numpy.mean(numpy.abs(derived2), axis=0))
                        else:
                            aggrDesc['dmean2'] = 'undefined'

                    if 'dvar2' in stats:
                        if not derived:
                            derived = [a - b for a, b in izip(values[1:], values[:-1])]
                        if not derived2:
                            derived2 = [a - b for a, b in izip(derived[1:], derived[:-1])]
                        if derived2:
                            aggrDesc['dvar2'] = essentia.array(numpy.var(derived2, axis=0))
                        else:
                            aggrDesc['dvar2'] = 'undefined'


                    if 'frames' in stats:
                        aggrDesc['frames'] = essentia.array(values)

                    if 'single_gaussian' in stats:
                        single_gaussian = essentia.SingleGaussian()
                        (m, cov, icov) = single_gaussian(essentia.array(values))
                        aggrDesc['mean'] = m
                        aggrDesc['cov'] = cov
                        aggrDesc['icov'] = icov

                    for stat in stats:
                        if stat.startswith('percentile_'):
                            p = float(stat.split('_')[1])
                            aggrDesc[stat] = essentia.array(percentile(values, p))


                except (TypeError, ValueError):  # values are not numeric

                    if len(values) == 1:
                       aggrDesc['value'] = values[0]
                    else:
                       aggrDesc['value'] = []
                       for value in values:
                           aggrDesc['value'].append(value)

        return aggregated



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

import numpy
from six import iteritems
from . import _essentia


# force the array objects to be of type float32
def array(object, **kwargs):
    return numpy.array(object, dtype='f4', **kwargs)


def zeros(object, **kwargs):
    return numpy.zeros(object, dtype='f4', **kwargs)


def ones(object, **kwargs):
    return numpy.ones(object, dtype='f4', **kwargs)


algoDecorator = lambda x: x


# An object representing an enum which contains int representations for
# essentia types. The purpose of this int representation is to have a common
# space for which to compare python and c++ types that are relevant to Essentia
class Edt:  # Essentia Data Type
    # c++ types
    BOOL = 'BOOL'
    INTEGER = 'INTEGER'
    REAL = 'REAL'
    STRING = 'STRING'
    STEREOSAMPLE = 'STEREOSAMPLE'
    COMPLEX = 'COMPLEX'
    VECTOR_INTEGER = 'VECTOR_INTEGER'
    VECTOR_REAL = 'VECTOR_REAL'
    VECTOR_STRING = 'VECTOR_STRING'
    VECTOR_COMPLEX = 'VECTOR_COMPLEX'
    VECTOR_VECTOR_STRING = 'VECTOR_VECTOR_STRING'
    VECTOR_VECTOR_REAL = 'VECTOR_VECTOR_REAL'
    VECTOR_VECTOR_COMPLEX = 'VECTOR_VECTOR_COMPLEX'
    VECTOR_STEREOSAMPLE = 'VECTOR_STEREOSAMPLE'
    MATRIX_REAL = 'MATRIX_REAL'
    MATRIX_COMPLEX = 'MATRIX_COMPLEX'
    VECTOR_MATRIX_REAL = 'VECTOR_MATRIX_REAL'
    VECTOR_TENSOR_REAL = 'VECTOR_TENSOR_REAL'
    TENSOR_REAL = 'TENSOR_REAL'
    POOL = 'POOL'

    # intermediate types
    LIST_EMPTY = 'LIST_EMPTY'
    LIST_MIXED = 'LIST_MIXED'
    LIST_INTEGER = 'LIST_INTEGER'
    LIST_REAL = 'LIST_REAL'
    LIST_LIST_REAL = 'LIST_LIST_REAL'
    LIST_LIST_INTEGER = 'LIST_LIST_INTEGER'
    LIST_LIST_EMPTY = 'LIST_LIST_EMPTY'
    LIST_COMPLEX = 'LIST_COMPLEX'
    LIST_LIST_COMPLEX = 'LIST_LIST_COMPLEX'
    LIST_ARRAY_REAL = 'LIST_ARRAY_REAL'
    LIST_ARRAY = 'LIST_ARRAY'
    NUMPY_FLOAT = 'NUMPY_FLOAT'
    UNDEFINED = 'UNDEFINED'

    def __init__(self, tp):
        self._tp = tp

    def isIntermediate(self):
        return self._tp in (Edt.LIST_EMPTY, Edt.LIST_MIXED, Edt.LIST_INTEGER,
                            Edt.LIST_REAL, Edt.LIST_LIST_REAL,
                            Edt.LIST_LIST_INTEGER, Edt.LIST_ARRAY,
                            Edt.LIST_ARRAY_REAL, Edt.UNDEFINED, Edt.NUMPY_FLOAT,
                            Edt.LIST_LIST_EMPTY, Edt.LIST_LIST_COMPLEX)

    def vectorize(self):
        return Edt('VECTOR_'+self._tp)

    def devectorize(self):
        if not self._tp.startswith('VECTOR_'):
            raise ValueError('cannot devectorize an Edt that is not a vector (Edt='+self._tp+')')

        return Edt(self._tp[len('VECTOR_'):])

    def __eq__(self, other):
        if isinstance(other, str):
            return self._tp == other
        elif isinstance(other, Edt):
            return self._tp == other._tp
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self._tp


# Determines the essentia data type of a given python object
def determineEdt(obj):
    # lists
    if isinstance(obj, list):
        if len(obj) == 0:
            return Edt(Edt.LIST_EMPTY)

        firstElmtType = determineEdt(obj[0])

        for item in obj[1:]:
            if determineEdt(item) != firstElmtType:
                return Edt(Edt.LIST_MIXED)

        if firstElmtType == Edt.INTEGER:
            return Edt(Edt.LIST_INTEGER)

        if firstElmtType == Edt.REAL:
            return Edt(Edt.LIST_REAL)

        if firstElmtType == Edt.STRING:
            return Edt(Edt.VECTOR_STRING)

        if firstElmtType == Edt.COMPLEX:
            return Edt(Edt.LIST_COMPLEX)

        if firstElmtType == Edt.VECTOR_STRING:
            return Edt(Edt.VECTOR_VECTOR_STRING)

        if firstElmtType == Edt.STEREOSAMPLE:
            raise TypeError('received a list of StereoSamples, if a VECTOR_STEREOSAMPLE is desired, create a 2-dimensional numpy.array of dimensions Nx2')

        if firstElmtType == Edt.MATRIX_REAL:
            return Edt(Edt.VECTOR_MATRIX_REAL)

        if firstElmtType == Edt.TENSOR_REAL:
            return Edt(Edt.VECTOR_TENSOR_REAL)

        if firstElmtType == Edt.LIST_REAL:
            return Edt(Edt.LIST_LIST_REAL)

        if firstElmtType == Edt.LIST_COMPLEX:
            return Edt(Edt.LIST_LIST_COMPLEX)

        if firstElmtType == Edt.LIST_INTEGER:
            return Edt(Edt.LIST_LIST_INTEGER)

        if firstElmtType == Edt.LIST_EMPTY:
            return Edt(Edt.LIST_LIST_EMPTY)

        if isinstance(obj[0], numpy.ndarray) and obj[0].ndim == 1:
            if obj[0].dtype == numpy.dtype('single'):
                return Edt(Edt.LIST_ARRAY_REAL)
            else:
                return Edt(Edt.LIST_ARRAY)

    if isinstance(obj, numpy.ndarray) and obj.ndim == 4:
        if obj.dtype == numpy.dtype('single'):
            return Edt(Edt.TENSOR_REAL)

        if obj.dtype == numpy.dtype('complex64'):
            return Edt(Edt.TENSOR_COMPLEX)

        raise TypeError('essentia can currently only accept two-dimensional numpy arrays of dtype '\
                        '"single"')

    # numpy array matrices
    if isinstance(obj, numpy.ndarray) and obj.ndim == 2:
        if obj.dtype == numpy.dtype('single'):
            return Edt(Edt.MATRIX_REAL)

        if obj.dtype == numpy.dtype('complex64'):
            return Edt(Edt.MATRIX_COMPLEX)

        raise TypeError('essentia can currently only accept two-dimensional numpy arrays of dtype '\
                        '"single"')

    # numpy arrays
    if isinstance(obj, numpy.ndarray) and obj.ndim == 1:
        if obj.dtype == numpy.dtype('single'):
            return Edt(Edt.VECTOR_REAL)
        if obj.dtype == numpy.dtype('int'):
            return Edt(Edt.VECTOR_INTEGER)
        if obj.dtype == numpy.dtype('complex64'):
            return Edt(Edt.VECTOR_COMPLEX)

        raise TypeError('essentia can currently only accept one-dimensional numpy arrays of dtype '\
                        '{"single", "int", "complex64"}, please consider using essentia.array to '\
                        'create your arrays')

    # bool (must go before ints! because True and False can be ints)
    if isinstance(obj, bool):
        return Edt(Edt.BOOL)

    # ints
    if isinstance(obj, int):
        return Edt(Edt.INTEGER)

    # reals
    if isinstance(obj, float):
        return Edt(Edt.REAL)

    # strings
    if isinstance(obj, str):
        return Edt(Edt.STRING)

    if isinstance(obj, numpy.complex64):
        return Edt(Edt.COMPLEX)

    if isinstance(obj, complex):
        return Edt(Edt.COMPLEX)

    if isinstance(obj, numpy.float32):
        return Edt(Edt.NUMPY_FLOAT)

    if isinstance(obj, dict):
        # map parameters
        if len(obj) > 0:
            firstType = None
            allKeysAreStrings = True
            allTypesEqual = True
            for key, val in iteritems(obj):
                if not isinstance(key, str):
                    allKeysAreStrings = False
                    break

                if firstType is None:
                    firstType = determineEdt(val)

                elif firstType != determineEdt(val):
                    allTypesEqual = False
                    break

            if allKeysAreStrings and allTypesEqual:
                return Edt('MAP_'+str(firstType))

    # pools
    if isinstance(obj, Pool) or isinstance(obj, _essentia.Pool):
        return Edt(Edt.POOL)

    # tuples
    if isinstance(obj, tuple) and len(obj) == 2 and \
       (isinstance(obj[0], float) or isinstance(obj[0], int)) and \
       (isinstance(obj[1], float) or isinstance(obj[1], int)):
        return Edt(Edt.STEREOSAMPLE)

    # everything else
    return Edt(Edt.UNDEFINED)


# Converts 'data' to 'goalType'. 'goalType' must be a non-intermediate EDT. If
# a conversion cannot be made, a TypeError will be raised.
def convertData(data, goalType):
    origType = determineEdt(data)

    if origType == goalType:
        return data

    if goalType == Edt.VECTOR_REAL:
        if origType == Edt.LIST_REAL or origType == Edt.LIST_INTEGER:
            return array(data)

        if origType == Edt.LIST_MIXED:
            for item in data:
                itemType = determineEdt(item)
                if not (itemType == Edt.REAL or itemType == Edt.INTEGER):
                    raise TypeError('Cannot convert data from type LIST_MIXED to type VECTOR_REAL ' +
                                    'because LIST_MIXED contains items not of type REAL or INTEGER (e.g. ' +
                                    str(itemType)+')')
            return array(data)

    if goalType == Edt.VECTOR_INTEGER:
        if origType == Edt.LIST_INTEGER:
            return numpy.array(data, numpy.int32)

    if goalType == Edt.VECTOR_REAL:
        if origType == Edt.LIST_INTEGER or\
           origType == Edt.VECTOR_INTEGER:
            return numpy.array(data, numpy.float32)

    if origType == Edt.LIST_EMPTY:
        if goalType == Edt.VECTOR_REAL:
            return array(data)
        if goalType == Edt.VECTOR_STRING:
            return data
        if goalType == Edt.VECTOR_INTEGER:
            return numpy.array(data, numpy.dtype('int'))
        if goalType == Edt.VECTOR_STEREOSAMPLE:
            return data

    if goalType == Edt.MATRIX_REAL and \
       (origType == Edt.LIST_LIST_REAL or origType == Edt.LIST_LIST_INTEGER or origType == Edt.LIST_ARRAY_REAL):
        return array(data)

    if goalType == Edt.VECTOR_VECTOR_REAL:
        if origType == Edt.MATRIX_REAL or origType == Edt.LIST_LIST_INTEGER:
            return [[float(col) for col in row] for row in data]

        if origType == Edt.LIST_LIST_REAL or origType == Edt.LIST_LIST_EMPTY or Edt.LIST_ARRAY_REAL:
            return data

    if goalType == Edt.REAL:
        if origType == Edt.INTEGER or origType == Edt.NUMPY_FLOAT:
            return float(data)

    if goalType == Edt.VECTOR_STEREOSAMPLE:
        if origType == Edt.LIST_LIST_INTEGER:
            for row in data:
                if len(row) != 2:
                    ValueError('Cannot convert a LIST_LIST_INTEGER to a VECTOR_STEREOSAMPLE if the sub-lists are not of length 2')
            return array(data)

        if origType == Edt.MATRIX_REAL:
            if data.shape[1] != 2:
                raise TypeError('Cannot convert a MATRIX_REAL to a VECTOR_STEREOSAMPLE if the second dimension of the MATRIX_REAL is not 2')

            return data

        if origType == Edt.LIST_MIXED:
            for row in data:
                if len(row) != 2:
                    ValueError('Cannot convert a LIST_MIXED to a VECTOR_STEREOSAMPLE if the sub-lists are not of length 2')
                try:
                    convertData(row, Edt.VECTOR_REAL)
                except:
                    TypeError('Cannot convert a LIST_MIXED to a VECTOR_STEREOSAMPLE if the sub-lists are not convertible to VECTOR_REAL')

            return array(data)

    if goalType == Edt.VECTOR_VECTOR_COMPLEX:
        if origType  == Edt.LIST_LIST_COMPLEX:
            return data

        if origType  == Edt.MATRIX_COMPLEX:
            return [[col for col in row] for row in data]

    raise TypeError('Cannot convert data from type %s (%s) to type %s' %
                    (str(origType), str(type(data)), str(goalType)))


class Pool:
    def __init__(self, poolRep=None):
        if poolRep is None:
            self.cppPool = _essentia.Pool()

        elif isinstance(poolRep, _essentia.Pool):
            self.cppPool = poolRep

        elif isinstance(poolRep, dict):
            self.cppPool = _essentia.Pool()
            for key, val in iteritems(poolRep):
                for v in val:
                    self.add(key, v)

        else:
            raise TypeError('poolRep argument must be a Cpp Essentia Pool or python dictionary')

    def add(self, key, value, validityCheck=False):
        givenType = determineEdt(value)

        # if we've seen this key before, determine the type
        if self.containsKey(key):
            goalType = Edt(self.cppPool.__keyType__(key)).devectorize()

        # if we haven't seen this type before, we will have to guess its type
        else:
            if givenType in (Edt.REAL, Edt.STRING, Edt.STEREOSAMPLE,
                             Edt.VECTOR_REAL, Edt.VECTOR_STRING,
                             Edt.VECTOR_STEREOSAMPLE, Edt.MATRIX_REAL,
                             Edt.TENSOR_REAL):
                goalType = givenType

            # some exceptions
            elif givenType == Edt.INTEGER:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.NUMPY_FLOAT:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.LIST_REAL:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.LIST_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.VECTOR_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)

            else:
                raise TypeError('Pool.add does not support the type: '+str(givenType))

        # try to convert the type
        try:
            convertedVal = convertData(value, goalType)
        except TypeError:
            raise KeyError('Pool.add could not convert given data to the type already in the Pool under the key \''+key+'\'')

        self.cppPool.__add__(key, str(goalType), convertedVal, validityCheck)

    def set(self, key, value, validityCheck=False):
        givenType = determineEdt(value)

        # if we've seen this key before, determine the type
        if self.containsKey(key):
            goalType = self.cppPool.__keyType__(key)

        # if we haven't seen this type before, we will have to guess its type
        else:
            if givenType in (Edt.REAL, Edt.STRING, Edt.VECTOR_REAL, Edt.TENSOR_REAL):
                goalType = givenType

            # some exceptions
            elif givenType == Edt.INTEGER:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.NUMPY_FLOAT:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.VECTOR_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.LIST_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.LIST_REAL:
                goalType = Edt(Edt.VECTOR_REAL)

            else:
                raise TypeError('Pool.set does not support the type: '+str(givenType))

        # try to convert the type
        try:
            convertedVal = convertData(value, goalType)
        except TypeError:
            raise KeyError('Pool.set could not convert given data to the type already in the Pool under the key \''+key+'\'')

        return self.cppPool.__set__(key, str(goalType), convertedVal, validityCheck)

    def merge(self, arg1, arg2=None, arg3=None):
        # this function is used to merge both descriptors and pools
        if determineEdt(arg1) == Edt.POOL:
            if arg3:
                raise TypeError('Pool.merge requires only 3 arguments when ' +
                                'merging entire pools: self, pool, string')
            if not arg2:
                mergeType = ''
            else:
                mergeType = arg2
            return self.cppPool.__merge__(str(Edt(Edt.POOL)), arg1.cppPool, mergeType)

        key = arg1
        value = arg2
        mergeType = arg3
        if not mergeType:
            mergeType = ''

        givenType = determineEdt(value)

        # if we've seen this key before, determine the type
        if self.containsKey(key):
            goalType = Edt(self.cppPool.__keyType__(key))  #.devectorize()

        # if we haven't seen this type before, we will have to guess its type
        else:
            if givenType in (Edt.REAL, Edt.STRING, Edt.STEREOSAMPLE,
                             Edt.VECTOR_REAL, Edt.VECTOR_STRING,
                             Edt.VECTOR_STEREOSAMPLE, Edt.MATRIX_REAL,
                             Edt.VECTOR_VECTOR_REAL, Edt.VECTOR_MATRIX_REAL,
                             Edt.VECTOR_TENSOR_REAL, Edt.TENSOR_REAL):
                goalType = givenType

            # some exceptions
            elif givenType == Edt.INTEGER:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.LIST_REAL:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.LIST_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.VECTOR_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)

            else:
                raise TypeError('Pool.merge does not support the type: '+str(givenType))

        # try to convert the type
        try:
            convertedVal = convertData(value, goalType)
        except TypeError:
            raise KeyError('Pool.merge could not convert given data to the type already in the Pool under the key\''+key+'\'')

        return self.cppPool.__merge__(key, str(goalType), convertedVal, mergeType)

    def mergeSingle(self, key, value, mergeType=''):
        givenType = determineEdt(value)

        # if we've seen this key before, determine the type
        if self.containsKey(key):
            goalType = self.cppPool.__keyType__(key)

        # if we haven't seen this type before, we will have to guess its type
        else:
            if givenType in (Edt.REAL, Edt.STRING, Edt.VECTOR_REAL):
                goalType = givenType

            # some exceptions
            elif givenType == Edt.INTEGER:
                goalType = Edt(Edt.REAL)
            elif givenType == Edt.VECTOR_INTEGER:
                goalType = Edt(Edt.VECTOR_REAL)
            elif givenType == Edt.NUMPY_FLOAT:
                goalType = Edt(Edt.REAL)

            else:
                raise TypeError('Pool.mergeSingle does not support the type: '+str(givenType))

        # try to convert the type
        try:
            convertedVal = convertData(value, goalType)
        except TypeError:
            raise KeyError('Pool.mergeSingle could not convert given data to the type already in the Pool under the key \''+key+'\'')

        return self.cppPool.__mergeSingle__(key, str(goalType), convertedVal, mergeType)

    def __getitem__(self, key):
        if not self.containsKey(key):
            raise KeyError('no key found named \''+key+'\'')

        return self.cppPool.__value__(key, self.cppPool.__keyType__(key))

    def containsKey(self, key):
        return key in self.descriptorNames()

    def descriptorNames(self, key=None):
        if not key:
            return self.cppPool.descriptorNames()
        return self.cppPool.descriptorNames(key)

    def remove(self, key):
        return self.cppPool.remove(key)

    def removeNamespace(self, namespace):
        return self.cppPool.removeNamespace(namespace)

    def isSingleValue(self, name):
        return self.cppPool.isSingleValue(name)

    def clear(self):
        return self.cppPool.clear()



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from six import iteritems
from . import _essentia
import essentia
import sys as _sys
from . import common as _c
from ._essentia import skeys as algorithmNames, sinfo as algorithmInfo

# Used as a place-holder for sources and sinks, implements the right shift
# operator
class _StreamConnector:

    def __init__(self, input_algo, output_algo, name):
        '''
        input_algo - is the algo to be used if StreamConnector is interpreted as a sink (right)
        output_algo - is the algo to be used if StreamConnector is interpreted as a source (left)
        name - is the name of the sink/source

        input_algo or output_algo should be set to None if StreamConnector cannot be interpreted as
        an input or output, respectively.
        '''
        self.input_algo = input_algo
        self.output_algo = output_algo
        self.name = name

    def __rshift__(left, right):
        if not isinstance(left.output_algo, VectorInput) and \
           not left.output_algo.hasOutput(left.name):
            raise NameError('The \'%s\' algorithm does not have a source called \'%s\''
                            %(left.output_algo.name(), left.name))

        # connect a source to a sink
        if isinstance(right, _StreamConnector):
            # need to make an exception if the left algorithm is a VectorInput
            # because its underlying c++ type is initialized at connection time
            if isinstance(left.output_algo, VectorInput):
                left.output_algo.__inner_init__(_c.Edt(right.input_algo.getInputType(right.name)))

            if not right.input_algo.hasInput(right.name):
                raise NameError('The \'%s\' algorithm does not have a sink called \'%s\''
                                %(right.input_algo.name(), right.name))

            _essentia.connect(left.output_algo, left.name,
                              right.input_algo, right.name)

            # update connections
            left.output_algo.connections[left].append(right)

            return right


        elif isinstance(right, _essentia.StreamingAlgorithm) and right.name() == 'FileOutput':
            _essentia.fileOutputConnect(left.output_algo, left.name, right)

            # update connections
            left.output_algo.connections[left].append(right)

            return right


        # connect a source to a pool
        elif isinstance(right, tuple):
            if not len(right) == 2 or \
               not isinstance(right[0], _c.Pool) or \
               not isinstance(right[1], str):
                raise TypeError('the right side tuple given should consist of a Pool and a string descriptor name')

            # make sure to lazy-initialize the VectorInput
            if isinstance(left.output_algo, VectorInput):
                left.output_algo.__inner_init_default__()

            # update connections
            left.output_algo.connections[left].append(right)

            return _essentia.poolConnect(left.output_algo, left.name, right[0].cppPool, right[1])

        # connect a source to NOWHERE
        elif right is None:
            # still need to lazy-initialize the VectorInput
            if isinstance(left.output_algo, VectorInput):
                left.output_algo.__inner_init_default__()

            # update connections
            left.output_algo.connections[left].append(right)

            return _essentia.nowhereConnect(left.output_algo, left.name)

        # none of the above accepted types: raise an exception
        raise TypeError('\'%s.%s\' A source can only be connected to a sink, a pair '\
                        '(tuple) of pool and key name, or None'
                        %(left.output_algo.name(), left.name))

    def disconnect(self, connector):
        # if they are not connected just return
        if connector not in self.output_algo.connections[self]: return;
        # remove connector from connections
        self.output_algo.connections[self].remove(connector)

        # call internal disconnect based on connector type
        if isinstance(connector, _StreamConnector):
            return _essentia.disconnect(self.output_algo, self.name, connector.input_algo, connector.name)

        elif isinstance(connector, tuple):
            if not len(connector) == 2 or \
               not isinstance(connector[0], _c.Pool) or \
               not isinstance(connector[1], str):
                raise TypeError('the tuple given should consist of a Pool and a string descriptor name')

            return _essentia.poolDisconnect(self.output_algo, self.name, connector[0].cppPool, connector[1])

        elif connector == None:
            return _essentia.nowhereDisconnect(self.output_algo, self.name)

        raise TypeError('\'%s.%s\' An output can only be disconnected from an '\
                        'input, a pair (tuple) of pool and key name, or None'
                        %(self.output_algo.name(), self.name))

    def totalProduced(self):
        return _essentia.totalProduced(self.output_algo, self.name)



def _create_streaming_algo(givenname):
    essentia.log.debug(essentia.EPython, 'Creating essentia.streaming class: %s' % givenname)

    _algoInstance = _essentia.StreamingAlgorithm(givenname)
    _algoDoc = _algoInstance.getDoc()
    _algoStruct = _algoInstance.getStruct()
    del _algoInstance

    class StreamingAlgo(_essentia.StreamingAlgorithm):
        __doc__ = _algoDoc
        __struct__ = _algoStruct

        def __init__(self, **kwargs):
            _essentia.StreamingAlgorithm.__init__(self, givenname)

            self.configure(**kwargs)

            # keys should be StreamConnectors (outputs) and values should be lists
            # of StreamConnectors/pool tuples/None (inputs)
            self.connections = {}

            # populate instance members from sources and sinks
            # we don't descriminate b/w inputs and outputs at this point, put
            # into a set (i.e. no duplicates)
            names = set(self.inputNames())
            for outputName in self.outputNames():
                names.add(outputName)

            for n in names:
                # TODO self should only be used in the first arg if this StreamConnector should be
                # an input, and in the second arg position if this StreamConnector should be an
                # output. Setting both to self won't hurt though, afaik.
                conn = _StreamConnector(self, self, n)
                self.connections[conn] = []
                setattr(self, n, conn)

        def configure(self, **kwargs):
            # verify that all types match and do any necessary conversions
            for name, val in iteritems(kwargs):
                goalType = self.paramType(name)
                try:
                    convertedVal = _c.convertData(val, goalType)
                except TypeError as e:
                    raise TypeError('Error verifying \''+name+'\' parameter: '+str(e))

                kwargs[name] = convertedVal

            # we need to keep a reference to these parameters so that they won't get cleaned up for
            # the lifetime of this algo
            self.__pyparams__ = kwargs

            self.__configure__(**kwargs)

    algoClass = _c.algoDecorator(StreamingAlgo)
    setattr(_sys.modules[__name__], givenname, algoClass)


# load all streaming algorithms into module
def _reloadStreamingAlgorithms():
    for name in algorithmNames():
        _create_streaming_algo(name)

_reloadStreamingAlgorithms()

# This subclass provides some more functionality for VectorInput
class VectorInput(_essentia.VectorInput):
    __doc__ = 'VectorInput v1.0\n\n\n'+\
              '  Outputs:\n\n'+\
              '    [variable] data - the given data\n\n\n'+\
              '  Description:\n\n'+\
              '    Can be used as the starting point of a streaming network. Its constructor\n'+\
              '    takes in a vector that is to be streamed one token at a time.'
    __struct__ = {'category': 'Input/output',
                  'description': 'Can be used as the starting point of a streaming network. Its constructor\ntakes in a vector that is to be streamed one token at a time.',
                  'inputs': [],
                  'name': 'VectorInput',
                  'outputs': [{'description': 'the given data',
                               'name': 'data',
                               'type': 'vector_<real>'}],
                  'parameters': []}


    def __init__(self, data):
        self.__initialized = False
        self.__initializedType = _c.Edt(_c.Edt.UNDEFINED)
        self.dataref = data
        self.data = _StreamConnector(None, self, 'data')

        # keys should be StreamConnectors (outputs) and values should be lists
        # of StreamConnectors/pool tuples/None (inputs)
        self.connections = {}

        self.connections[self.data] = []

    # initializes the underlying c++ type
    # sinkEdt represents the type that the VectorInput should be initialized to
    # (e.g. the type of the sink we're connecting this to)
    def __inner_init__(self, sinkEdt):
        # get vector form of sinkEdt, because we'll compare to given data input
        # which is a vector of something
        sinkEdt = sinkEdt.vectorize()

        if self.__initialized:
            if self.__initializedType == sinkEdt:
                return
            raise TypeError('VectorInput was already connected to another sink with a different type, original type: '+str(self.__initializedType)+' new type: '+str(sinkEdt))

        self.dataref = _c.convertData(self.dataref, sinkEdt)
        _essentia.VectorInput.__init__(self, self.dataref, str(sinkEdt))
        self.__initializedType = sinkEdt #_c.Edt
        self.__initialized = True


    # This method should be used if the sink being connected to is also
    # ambiguous (e.g. Pool or NOWHERE). This method will guess the type of the
    # VectorInput (a type that would be valid for a pool (e.g. VECTOR_REAL,
    # VECTOR_STRING)) and raise a TypeError if a type could not be guessed.
    def __inner_init_default__(self):
        if self.__initialized:
            return

        sourceEdt = _c.determineEdt(self.dataref)

        if not sourceEdt.isIntermediate():
            _essentia.VectorInput.__init__(self, self.dataref, str(sourceEdt))
            self.__initialized = True
            return

        if sourceEdt == _c.Edt.LIST_EMPTY or \
           sourceEdt == _c.Edt.LIST_INTEGER or \
           sourceEdt == _c.Edt.LIST_REAL or \
           sourceEdt == _c.Edt.LIST_MIXED:
            self.dataref = _c.convertData(self.dataref, _c.Edt.VECTOR_REAL)
            _essentia.VectorInput.__init__(self, self.dataref, _c.Edt.VECTOR_REAL)
            self.__initialized = True
            return

        if sourceEdt == _c.Edt.LIST_LIST_REAL or sourceEdt == _c.Edt.LIST_LIST_INTEGER:
            self.dataref = _c.convertData(self.dataref, _c.Edt.MATRIX_REAL)
            _essentia.VectorInput.__init__(self, self.dataref, _c.Edt.MATRIX_REAL)
            self.__initialized = True
            return

        raise TypeError('Unable to initialize VectorInput because it is '+\
                        'being connected to None or a Pool, and the '+\
                        'VectorInput\'s data consists of an unsupported Pool '+\
                        'type: '+str(sourceEdt))

class CompositeBase(object):
    '''
    Inherit from this class when creating a new composite streaming algorithm.
    '''

    def __init__(self):
        self.inputs = {}
        self.outputs = {}

    def name(self): return self.__class__.__name__

    def hasInput(self, name):
        return name in self.inputs

    def hasOutput(self, name):
        return name in self.outputs

    def __getattr__(self, name):
        '''
        returns a _StreamConnector that can be used for connecting to inputs/outputs
        '''
        # just return the streamconnector of the real algorithm which this
        # streamconnector is pointing to:
        if self.hasInput(name) and  self.hasOutput(name):
            raise NameError('CompositeBase algorithms cannot '\
                            'have a sink and a source with the same '\
                            'name:\''+self.name()+'.'+name+'\'')
        elif self.hasInput(name):  return self.inputs[name]
        elif self.hasOutput(name): return self.outputs[name]
        raise NameError('The '+self.name()+ ' CompositeBase algorithm does not have a connector named \''+name+'\'')



# Copyright (C) 2006-2016  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

#!/usr/bin/python

from __future__ import print_function

import os
import sys
import glob
import io


def get_all_algorithms(algo_dir, root_dir=None):
    algorithms = {}

    for root, dirs, files in os.walk(algo_dir):
        for filename in glob.glob(os.path.join(root, '*.h')):

            try:
                lines = [line.strip() for line in io.open(filename, encoding='utf8')]

                # get a nicer filename now that it's been read
                if filename.startswith(root_dir):
                    filename = filename[len(root_dir):]
                    if filename[0] == '/' or filename[0] == '\\':  # both linux and win
                        filename = filename[1:]

                algo = ''
                salgo = ''  # name of streaming algo
                has_standard = False
                has_streaming = False
                parameters = {}
                inputs = {}
                outputs = {}

                for line in lines:
                    if line.startswith('//'):
                        continue

                    if line.find('namespace standard') >= 0:
                        has_standard = True
                        continue

                    if line.find('namespace streaming') >= 0:
                        has_streaming = True
                        continue

                    if (line.find('public Algorithm') >= 0 or
                        line.find('public AccumulatorAlgorithm') >= 0 or
                        line.find('public StreamingAlgorithmWrapper') >= 0):
                        name = line.split(' ')[1]

                        if has_standard and not has_streaming: algo = name

                        if not has_standard and has_streaming: salgo = name

                        if has_standard and has_streaming:
                            if algo == '':
                                algo = name
                            else:
                                salgo = name

                        # if we have both standard and streaming, stop here because we
                        # don't want to override the previous values
                        if has_standard and has_streaming:
                            break

                    if line.find('Input<') >= 0:
                        line = line[line.find('Input<')+6:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(';', '').strip()
                        inputs[var_name] = {'type': var_type}
                        continue

                    if line.find('Output<') >= 0:
                        line = line[line.find('Output<')+7:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(";", "").strip()
                        outputs[var_name] = {'type': var_type}
                        continue

                    if line.find('Sink<') >= 0:
                        line = line[line.find('Sink<')+5:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(';', '').strip()
                        inputs[var_name] = {'type': var_type}
                        continue

                    if line.find('SinkProxy<') >= 0:
                        line = line[line.find('SinkProxy<')+10:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(';', '').strip()
                        inputs[var_name] = {'type': var_type}
                        continue

                    if line.find('Source<') >= 0:
                        line = line[line.find('Source<')+7:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(";", "").strip()
                        outputs[var_name] = {'type': var_type}
                        continue

                    if line.find('SourceProxy<') >= 0:
                        line = line[line.find('SourceProxy<')+12:]
                        var_type = line[:line.rfind('>')].strip()
                        var_name = line[line.rfind('>')+1:].replace(";", "").strip()
                        outputs[var_name] = {'type': var_type}
                        continue

                    if line.find('declareParameter(') >= 0:
                        line = line[line.find("(")+1:]
                        line = line[:line.find(");")]
                        name = line.split(',', 1)[0][1:-1]
                        line = line.split(',', 1)[1].lstrip(" \"")
                        description = line[0:line.find("\"")]
                        rest = line[line.find("\""):]
                        default = rest.replace("\", ", "")
                        parameters[name] = (description, default)
                        continue

                    if line.find('declareInput') >= 0:
                        line = line[line.find('(')+1:]
                        line = line[0:line.find(');')]
                        parts = line.strip().split(',', 3)
                        len_parts = len(parts)
                        if len_parts == 2:
                            parts.append("")

                        if len_parts == 4:
                            input_name_idx = 3
                            input_desc_idx = -1
                        elif len_parts < 4:
                            if not has_streaming:
                                input_name_idx = 1
                                input_desc_idx = 2
                            else:
                                input_name_idx = 2
                                input_desc_idx = -1
                        if input_name_idx >= 0:
                            input_name = parts[input_name_idx].strip(" ,\"")
                        else:
                            input_name = ''
                        if input_desc_idx >= 0:
                            input_desc = parts[input_desc_idx].strip(" ,\"")
                        else:
                            input_desc = input_name

                        if not has_streaming:
                            var_name = parts[0]
                        else:
                            var_name = parts[0]
                            # big hack to correctly parser stuff like declareInput(_algo->input("a"))
                            if var_name.find('-') >= 0:
                                inputs[var_name] = {'type': 'unknown'}
                        inputs[var_name]['name'] = input_name
                        inputs[var_name]['description'] = input_desc
                        continue

                    if line.find('declareOutput') >= 0:
                        #line = line[line.find('(')+1:]
                        #line = line[0:line.find(');')]
                        #parts = line.strip().split(',', 2) #[1:]
                        #if len(parts) == 2:
                        #    parts.append("")
                        #var_name = parts[0]
                        #output_name = parts[1].strip(" ,\"")
                        #output_desc = parts[2].strip(" ,\"")
                        #outputs[var_name]['name'] = output_name
                        #outputs[var_name]['description'] = output_desc
                        #continue

                        line = line[line.find('(')+1:]
                        line = line[0:line.find(');')]
                        parts = line.strip().split(',', 3)
                        len_parts = len(parts)
                        if len_parts == 2:
                            parts.append("")

                        if len_parts == 4:
                            input_name_idx = 3
                            input_desc_idx = -1
                        elif len_parts < 4:
                            if not has_streaming:
                                input_name_idx = 1
                                input_desc_idx = 2
                            else:
                                input_name_idx = 2
                                input_desc_idx = -1
                        if input_name_idx >= 0:
                            input_name = parts[input_name_idx].strip(" ,\"")
                        else:
                            input_name = ''
                        if input_desc_idx >= 0:
                            input_desc = parts[input_desc_idx].strip(" ,\"")
                        else:
                            input_desc = input_name

                        if not has_streaming:
                            var_name = parts[0]
                        else:
                            var_name = parts[0]
                            # big hack to correctly parser stuff like declareInput(_algo->input("a"))
                            if var_name.find('-') >= 0:
                                inputs[var_name] = {'type': 'unknown'}
                        
                        outputs[var_name]['name'] = input_name
                        outputs[var_name]['description'] = input_desc
                        continue

                # Sanity checks

                # 1- if a file is in the algo dir, it *must* be an algorithm (now a warning)
                if not has_standard and not has_streaming:
                    print('WARNING: file "%s" does not seem to contain an algorithm.' % filename)
                    continue
                    #sys.exit(1)

                # 2- if an algo has both standard & streaming form in the same file,
                #    they must have the same name
                if has_standard and has_streaming and algo != salgo:
                    print('ERROR: file "%s" contains both standard and streaming'\
                          'version of the algorithm, but they have different '\
                          'names: %s != %s.' %(filename, algo, salgo))
                    sys.exit(1)

                # 3- if algo is only available as streaming, use salgo as algo name
                if has_streaming:
                    algo = salgo

                # 4- if we couldn't determine some attributes of inputs/outputs, put
                #    placeholders
                for (input, attr) in inputs.items():
                    if 'name' not in attr:
                        attr['name'] = 'Unknown'
                    if 'description' not in attr:
                        attr['description'] = 'TODO'

                for (output, attr) in outputs.items():
                    if 'name' not in attr:
                        attr['name'] = 'Unknown'
                    if 'description' not in attr:
                        attr['description'] = 'TODO'

                algorithms[algo] = {'header': filename,
                                    'source': filename.replace('.h', '.cpp'),
                                    'has_standard': has_standard,
                                    'has_streaming': has_streaming,
                                    'inputs': inputs,
                                    'outputs': outputs,
                                    'parameters': parameters}
            except:
                print('Error while trying to parse file "%s"' % filename)
                raise

    return algorithms


def create_registration_cpp(all_algos, registration_filename, use_streaming=True):

    cpp_code = "#include \"algorithmfactory.h\"\n"

    # write #include's
    for algo in all_algos:
        if all_algos[algo]['has_standard'] or (use_streaming and all_algos[algo]['has_streaming']):
            # Essentia's Duration conflicts with Accelerates, so bump the include to the top
            if "FFTA" in algo:
                cpp_code = '#include "%s"\n' % all_algos[algo]['header'] + cpp_code
            else:
                cpp_code += '#include "%s"\n' % all_algos[algo]['header']

    # register standard algorithms in factory
    cpp_code += "\nnamespace essentia {\nnamespace standard {\n\nESSENTIA_API void registerAlgorithm() {\n"

    for algo in all_algos:
        if all_algos[algo]['has_standard']:
            cpp_code += "    AlgorithmFactory::Registrar<%s> reg%s;\n" % (algo, algo)

    cpp_code += "}}}\n"

    cpp_code += "\n"
    cpp_code += "\n"

    # register streaming algorithms in factory
    if use_streaming:
        cpp_code += "\nnamespace essentia {\nnamespace streaming {\n\nESSENTIA_API void registerAlgorithm() {\n"

        for algo in all_algos:
            if all_algos[algo]['has_streaming']:
                if all_algos[algo]['has_standard']:
                    cpp_code += "    AlgorithmFactory::Registrar<%s, essentia::standard::%s> reg%s;\n" % (algo, algo, algo)
                else:
                    cpp_code += "    AlgorithmFactory::Registrar<%s> reg%s;\n" % (algo, algo)

        cpp_code += "}}}\n"

    with open(registration_filename, "w") as f:
        f.write(cpp_code)


def create_version_h(version_filename, essentia_version, git_sha):
    h_code = """
#ifndef VERSION_H_
#define VERSION_H_
#define ESSENTIA_VERSION "%s"
#define ESSENTIA_GIT_SHA "%s"
#endif /* VERSION_H_ */
""" % (essentia_version, git_sha)

    with open(version_filename, "w") as f:
        f.write(h_code)


if __name__ == '__main__':
    algo_dir = sys.argv[1]
    algos = get_all_algorithms(algo_dir)

    for name, algo in algos.items():
        print('%s:' % name)
        for attr in algo:
            print('  %s: %s' % (attr, algo[attr]))
        print()



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

import numpy as np
import essentia.standard as es
from essentia import Pool, array


def nsgcqgram(audio, frameSize=8192, transitionSize=1024, minFrequency=65.41,
              maxFrequency=6000, binsPerOctave=48,
              sampleRate=44100, rasterize='full',
              phaseMode='global', gamma=0,
              normalize='none', window='hannnsgcq'):
    """Frame-wise invertible Constant-Q analysis.
    This code replicates the Sli-CQ algorithm from [1]. A Tukey window
    is used to perform a zero-phased, zero-padded, half-overlapped
    frame-wise analysis with the `NSGConstantQ` algorithm.

    References:
      [1] Velasco, G. A., Holighaus, N., Dörfler, M., & Grill, T. (2011).
        "Constructing an invertible constant-Q transform with non-stationary
        Gabor frames". Proceedings of DAFX11, Paris, 93-99.

    Args:
        audio (vector): If it is empty, an exception is raised.
    Returns:
        (list of 2D complex arrays): Time/frequency complex matrices representing the NSGCQ `constantq` coefficients for each `frameSize // 2` samples jump.
        (list of complex vectors): Complex vectors representing the NSGCQ `constantqdc` coefficients for each `frameSize // 2` samples jump.
        (list of complex vectors): Complex vectors representing the NSGCQ `constantqnf` coefficients for each `frameSize // 2` samples jump.
    """

    hopSize = frameSize // 2
    halfTransitionSize = transitionSize // 2

    NSGCQ = es.NSGConstantQ(inputSize=frameSize, minFrequency=minFrequency,
                            maxFrequency=maxFrequency, binsPerOctave=binsPerOctave,
                            sampleRate=sampleRate, rasterize=rasterize,
                            phaseMode=phaseMode, gamma=gamma,
                            normalize=normalize, window=window,
                            minimumWindow=8)

    w = es.Windowing(type='hannnsgcq', normalized=False, zeroPhase=False)(
        np.ones(transitionSize * 2).astype('float32'))
    window = np.hstack(
        [w[-transitionSize:], np.ones(hopSize - transitionSize), w[:transitionSize]])

    # Indexes for the windowing process.
    evenWin = np.hstack([np.arange(frameSize - halfTransitionSize,
                                   frameSize), np.arange(hopSize + halfTransitionSize)])
    oddWin = np.hstack([np.arange(hopSize - halfTransitionSize,
                                  frameSize), np.arange(halfTransitionSize)])
    h0 = np.arange(-halfTransitionSize, hopSize + halfTransitionSize)

    # Zero-pad the signal for an integer number of frames.
    zeroPad = (frameSize - window.size) // 2
    frameNum = int(np.ceil(audio.size / hopSize))

    x = np.hstack([audio, np.zeros(hopSize * frameNum -
                                   audio.size + hopSize)]).astype('float32')
    frames = np.zeros([frameNum, frameSize], dtype='float32')

    # Mirror-pad for the first frame.
    frames[0][evenWin] = np.hstack(
        [-x[halfTransitionSize - 1::-1], x[:hopSize + halfTransitionSize]]) * window

    # Slice audio.
    for kk in range(2, frameNum, 2):
        frames[kk][evenWin] = x[(kk - 1) * hopSize + h0] * window

    for kk in range(1, frameNum, 2):
        frames[kk][oddWin] = x[(kk - 1) * hopSize + h0] * window

    # Compute the transform.
    cqShift, dcShift, nfShift = [], [], []
    for i in range(frameNum):
        cqFrame, dcFrame, nfFrame = NSGCQ(frames[i])
        cqShift.append(cqFrame)
        dcShift.append(dcFrame)
        nfShift.append(nfFrame)

    cqSize = cqFrame.shape[1]
    dcSize = dcFrame.size
    nfSize = nfFrame.size

    # Center the frames for a better display.
    cqShiftEven = np.hstack(
        [np.arange(cqSize // 4, cqSize), np.arange(0, cqSize // 4)])
    dcShiftEven = np.hstack(
        [np.arange(dcSize // 4, dcSize), np.arange(0, dcSize // 4)])
    nfShiftEven = np.hstack(
        [np.arange(nfSize // 4, nfSize), np.arange(0, nfSize // 4)])

    cqShiftOdd = np.hstack(
        [np.arange(3 * cqSize // 4, cqSize), np.arange(0, 3 * cqSize // 4)])
    dcShiftOdd = np.hstack(
        [np.arange(3 * dcSize // 4, dcSize), np.arange(0, 3 * dcSize // 4)])
    nfShiftOdd = np.hstack(
        [np.arange(3 * nfSize // 4, nfSize), np.arange(0, 3 * nfSize // 4)])

    cq = [cqShift[i][:, cqShiftEven] if i % 2 else cqShift[i][:, cqShiftOdd]
          for i in range(len(cqShift))]
    dc = [dcShift[i][dcShiftEven] if i % 2 else dcShift[i][dcShiftOdd]
          for i in range(len(dcShift))]
    nf = [nfShift[i][nfShiftEven] if i % 2 else nfShift[i][nfShiftOdd]
          for i in range(len(nfShift))]

    return cq, dc, nf


def __inverseTukeyWindow__(x):
    return (1 + np.cos(np.pi * x)) / (1 + np.cos(np.pi * x) ** 2)


def nsgicqgram(cq, dc, nf, frameSize=8192, transitionSize=1024, minFrequency=65.41,
               maxFrequency=6000, binsPerOctave=48,
               sampleRate=44100, rasterize='full',
               phaseMode='global', gamma=0,
               normalize='none', window='hannnsgcq'):
    """Frame-wise invertible Constant-Q synthesis.
    This code replicates the Sli-CQ algorithm from [1]. An inverse Tukey window
    is used to resynthetise the original audio signal from the `nsgcqgram`
    representation using the `NSGIConstantQ` algorithm.

    References:
      [1] Velasco, G. A., Holighaus, N., Dörfler, M., & Grill, T. (2011).
        "Constructing an invertible constant-Q transform with non-stationary
        Gabor frames". Proceedings of DAFX11, Paris, 93-99.

    Args:
        (list of 2D complex arrays): Time / frequency complex matrices representing the NSGCQ `constantq` coefficients for each `frameSize // 2` samples jump.
        (list of complex vectors): Complex vectors representing the NSGCQ `constantqdc` coefficients for each `frameSize // 2` samples jump.
        (list of complex vectors): Complex vectors representing the NSGCQ `constantqnf` coefficients for each `frameSize // 2` samples jump.
    Returns:
        audio (vector): The synthetized audio.
    """

    hopSize = frameSize // 2
    halfTransitionSize = transitionSize // 2

    cqSize = cq[0].shape[1]
    dcSize = len(dc[0])
    nfSize = len(nf[0])

    NSGICQS = es.NSGIConstantQ(inputSize=frameSize, minFrequency=minFrequency,
                               maxFrequency=maxFrequency, binsPerOctave=binsPerOctave,
                               sampleRate=sampleRate, rasterize=rasterize,
                               phaseMode=phaseMode, gamma=gamma,
                               normalize=normalize, window=window,
                               minimumWindow=8)

    # Tukey inverse window.
    window = np.zeros(frameSize)
    window[np.arange((hopSize + transitionSize) // 2,
                     (3 * hopSize - transitionSize) // 2)
           ] = np.ones(hopSize - transitionSize)

    window[np.hstack([np.arange((hopSize - transitionSize) // 2,
                                (hopSize + transitionSize) // 2),
                      np.arange((3 * hopSize - transitionSize) // 2,
                                (3 * hopSize + transitionSize) // 2)])
           ] = __inverseTukeyWindow__(np.arange(-transitionSize,
                                                transitionSize) / transitionSize)

    # Undo the frame centering.
    cqShiftEven = np.hstack(
        [np.arange(3 * cqSize // 4, cqSize), np.arange(0, 3 * cqSize // 4)])
    dcShiftEven = np.hstack(
        [np.arange(3 * dcSize // 4, dcSize), np.arange(0, 3 * dcSize // 4)])
    nfShiftEven = np.hstack(
        [np.arange(3 * nfSize // 4, nfSize), np.arange(0, 3 * nfSize // 4)])

    cqShiftOdd = np.hstack(
        [np.arange(cqSize // 4, cqSize), np.arange(0, cqSize // 4)])
    dcShiftOdd = np.hstack(
        [np.arange(dcSize // 4, dcSize), np.arange(0, dcSize // 4)])
    nfShiftOdd = np.hstack(
        [np.arange(nfSize // 4, nfSize), np.arange(0, nfSize // 4)])

    cqShift = [cq[i][:, cqShiftEven] if i % 2 else cq[i][:, cqShiftOdd]
          for i in range(len(cq))]
    dcShift = [dc[i][dcShiftEven] if i % 2 else dc[i][dcShiftOdd]
          for i in range(len(dc))]
    nfShift = [nf[i][nfShiftEven] if i % 2 else nf[i][nfShiftOdd]
          for i in range(len(nf))]


    # Loop to store the audio frames.
    frames = [NSGICQS(cqShift[i], dcShift[i], nfShift[i])
              for i in range(len(cq))]


    # Overlap-add.
    audio = np.zeros((len(frames) + 1) * hopSize)

    for kk in range(len(frames)):
        audio[np.arange(int(np.ceil((kk - .5) * hopSize)),
                        int(np.ceil((kk + 1.5) * hopSize)))
              ] += np.roll(frames[kk], int(np.floor(((-1) ** kk) * hopSize / 2))) * window

    return audio[hopSize:]


def nsgcq_overlap_add(cq):
    """Frame-wise invertible Constant-Q synthesis.
    This function performs the overlap-add process of the CQ frames obtained by nsgcq_gram.
    The output of this algorithm may be used for visualization purposes.
    Note: It is not possible to perform a perfect reconstruction from the overlapped version of the CQ data. 

    Args:
        (list of 2D complex arrays): Time / frequency complex matrices representing the NSGCQ `constantq` coefficients for each `frameSize // 2` samples jump.
    Returns:
        (2D complex array): The overlapped version of the Constant-Q. 
    """

    frameNum = len(cq)
    cqChannels = cq[0].shape[0]
    cqSize = cq[0].shape[1]

    hopSize = cqSize // 2
    timeStamps = (frameNum + 1) * hopSize

    index = np.arange(cqSize)

    cqOverlap = np.zeros(
        [cqChannels, timeStamps], dtype='complex')


    # Overlap-add.
    for jj in range(frameNum):
        cqOverlap[:, jj * hopSize + index] += cq[jj]

    return cqOverlap[:, hopSize:]



def hpcpgram(audio, sampleRate=44100, frameSize=4096, hopSize=2048, numBins=12,
            windowType='blackmanharris62', minFrequency=100, maxFrequency=4000, 
            whitening=False, maxPeaks=100, magnitudeThreshold=1e-05, **kwargs):
    """
    Compute Harmonic Pitch Class Profile (HPCP) Grams for overlapped frames of a given input audio signal 

    For additional list of parameters of essentia standard mode HPCP please refer to 
    http://essentia.upf.edu/documentation/reference/std_HPCP.html

    References:
    [1]. Gómez, E. (2006). Tonal Description of Polyphonic Audio for Music Content Processing.

    Inputs
        audio (2d vector): audio signal

    Parameters:
        sampleRate : (real ∈ (0, ∞), default = 44100) :
        the sampling rate of the audio signal [Hz]

        frameSize (integer ∈ [1, ∞), default = 1024) :
        the output frame size
        
        hopSize (integer ∈ [1, ∞), default = 512) :
        the hop size between frames

        numBins : (integer ∈ [12, ∞), default = 12) :
        the size of the output HPCP (must be a positive nonzero multiple of 12)

        windowType (string ∈ {hamming, hann, hannnsgcq, triangular, square, blackmanharris62, blackmanharris70, blackmanharris74, blackmanharris92}, default = blackmanharris62) :
        the window type, which can be 'hamming', 'hann', 'triangular', 'square' or 'blackmanharrisXX'

        maxFrequency : (real ∈ (0, ∞), default = 4000) :
        the maximum frequency that contributes to the SpectralPeaks and HPCP algorithms computation [Hz] (the difference between the max and split frequencies must not be less than 200.0 Hz)

        minFrequency : (real ∈ (0, ∞), default = 100) :
        the minimum frequency that contributes to the SpectralPeaks and HPCP algorithm computation [Hz] (the difference between the min and split frequencies must not be less than 200.0 Hz)

        maxPeaks (integer ∈ [1, ∞), default = 100) :
        the maximum number of returned peaks while calculating SpectralPeaks

        magnitudeThreshold (real ∈ (-∞, ∞), default = 0) :
        peaks below this given threshold are not outputted while calculating Spectral Peaks

        whitening : (boolean (True, False), default = False)
        Optional step of computing spectral whitening to the output from speakPeak magnitudes

        kwargs : additional keyword arguments
        Arguments to parameterize HPCP alogithms.
        see standard mode HPCP algorithm (http://essentia.upf.edu/documentation/reference/std_HPCP.html).


    Returns: hpcpgram of overlapped frames of input audio signal (2D vector) 

    """
    frameGenerator = es.FrameGenerator(array(audio), frameSize=frameSize, hopSize=hopSize)
    window = es.Windowing(type=windowType)
    spectrum = es.Spectrum()
    # Refer http://essentia.upf.edu/documentation/reference/std_SpectralPeaks.html
    spectralPeaks = es.SpectralPeaks(magnitudeThreshold=magnitudeThreshold,
                                     maxFrequency=maxFrequency,
                                     minFrequency=minFrequency,
                                     maxPeaks=maxPeaks,
                                     sampleRate=sampleRate)
    # http://essentia.upf.edu/documentation/reference/std_SpectralWhitening.html
    spectralWhitening = es.SpectralWhitening(maxFrequency= maxFrequency,
                                            sampleRate=sampleRate)
    # http://essentia.upf.edu/documentation/reference/std_HPCP.html
    hpcp = es.HPCP(sampleRate=sampleRate,
                   maxFrequency=maxFrequency,
                   minFrequency=minFrequency,
                   size=numBins, **kwargs)
    pool = Pool()
    #compute hpcp for each frame and add the results to the pool
    for frame in frameGenerator:
        spectrum_mag = spectrum(window(frame))
        frequencies, magnitudes = spectralPeaks(spectrum_mag)
        if whitening:
            w_magnitudes = spectralWhitening(spectrum_mag,
                                            frequencies,
                                            magnitudes)
            hpcp_vector = hpcp(frequencies, w_magnitudes)
        else:
            hpcp_vector = hpcp(frequencies, magnitudes)
        pool.add('tonal.hpcp', hpcp_vector)
    return pool['tonal.hpcp']




# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

import inspect, types
import streaming
import _essentia
import common
from streaming import _reloadStreamingAlgorithms

# genetic marker used to track which composite parameters configure which inner algorithms
class MarkerObject(object):
    def __init__(self, default_value=None):
        self.default_value = default_value


edt_parameter_code = { common.Edt.STRING:             'String',
                       common.Edt.INTEGER:            'Int',
                       common.Edt.VECTOR_REAL:        'VectorReal',
                       common.Edt.VECTOR_VECTOR_REAL: 'VectorVectorReal',
                       common.Edt.REAL:               'Real',
                       common.Edt.BOOL:               'Bool'}


edt_cpp_code = { common.Edt.STRING:             'string',
                 common.Edt.INTEGER:            'int',
                 common.Edt.VECTOR_INTEGER:     'vector<int>',
                 common.Edt.VECTOR_STRING:      'vector<string>',
                 common.Edt.VECTOR_REAL:        'vector<Real>',
                 common.Edt.VECTOR_VECTOR_REAL: 'vector<vector<Real> >',
                 common.Edt.REAL:               'Real',
                 common.Edt.BOOL:               'bool'}


# html code for edt_dot_code:
# &lt; --> less than or <
# &gt; --> greater than or >
# #&#58; --> colon or :

edt_dot_code = { common.Edt.STRING:             'string',
                 common.Edt.INTEGER:            'int',
                 common.Edt.VECTOR_INTEGER:     'vector&lt;int&gt;',
                 common.Edt.VECTOR_STRING:      'vector&lt;string&gt;',
                 common.Edt.VECTOR_REAL:        'vector&lt;Real&gt;',
                 common.Edt.VECTOR_VECTOR_REAL: 'vector&lt;vector&lt;Real&gt; &gt;',
                 common.Edt.REAL:               'Real',
                 common.Edt.BOOL:               'bool',
                 common.Edt.STEREOSAMPLE:       'StereoSample',
                 common.Edt.MATRIX_REAL:        'TNT&#58;&#58;Array2D&lt;Real&gt;'} #&#58; -> colon

# finds the EDT of a parameter of an algorithm given the name of the parameter, the marker_obj used
# to configure that parameter, and the configure_log
def find_edt(composite_param_name, marker_obj, configure_log):
    # find inner algorithm and inner parameter name that this composite_param will configure
    for inner_algo_name, properties in configure_log.iteritems():
        for inner_param_name, value in properties['parameters'].iteritems():
            if marker_obj == value:
                return properties['instance'].paramType(inner_param_name)

    raise RuntimeError('Could not determine parameter type of composite algorithm\'s \''+composite_param_name+'\' parameter')


# given a reference to an inner algorithm and the configure_log, returns the name of the algo (use
# lower() if referring to the member var name)
def inner_algo_name(instance, configure_log):
    for algo_name, properties in configure_log.iteritems():
        if instance == properties['instance']:
            return algo_name

    raise RuntimeError('Could not find the name of the inner algorithm')


def generate_dot_algo(algo_name, algo_inst):
    '''declares an algorithm in dot'''
    dot_code = '_'+algo_name.lower()+' [shape="box", style="rounded,filled", fillcolor="grey50", color="transparent", \n'
    indent = ' '*len('_'+algo_name)
    dot_code += indent+'  label=<\n'
    indent += ' '*len('label=<')
    dot_code += generate_dot_algo_label(algo_inst, indent)
    dot_code += '>]\n\n'

    return dot_code


def generate_dot_algo_label(algo_inst, indent=''):
    ### each label of a node consists of algo inputs, algo name,
    ### configuration parameters and algo outputs

    dot_code =          indent+'<table border="0"><tr>\n'+\
                        indent+'    <td><table border="0" bgcolor="white">\n'

    # add inputs:
    if not len(algo_inst.inputNames()): dot_code += ''
    else:
        for name in algo_inst.inputNames():
            typestr = edt_dot_code[ algo_inst.getInputType(name) ]
            dot_code += indent+'        <tr><td port="'+name+'_i">'+name+'<br/>['+typestr+']</td></tr>\n'

    dot_code +=         indent+'    </table></td>\n\n'+\
                        indent+'    <td><table border="0">\n'

    # add algo name:
    dot_code +=         indent+'        <tr><td valign="top" colspan="2"><font color="white" point-size="18">'+algo_inst.name()+'</font></td></tr>\n'

    # add parameters:
    if not algo_inst.parameterNames(): dot_code += ''
    else:
        for name in algo_inst.parameterNames():
            value = algo_inst.paramValue(name)
            dot_code += indent+'        <tr>\n'+\
                        indent+'            <td border="0" valign="top" align="right"><font color="white">'+name+'</font></td>\n'+\
                        indent+'            <td border="0" valign="top" align="left"><font color="white">'+str(value)+'</font></td>\n'+\
                        indent+'        </tr>\n'

    dot_code +=         indent+'    </table></td>\n\n'+\
                        indent+'    <td><table border="0" bgcolor="white">\n'

    # add outputs:
    if not len(algo_inst.outputNames()): dot_code += ''
    else:
        for name in algo_inst.outputNames():
            typestr = edt_dot_code[ algo_inst.getOutputType(name) ]
            dot_code += indent+'        <tr><td port="'+name+'_o">'+name+'<br/>['+typestr+']</td></tr>\n'

    dot_code +=         indent+'    </table></td>\n'+\
                        indent+'</tr></table>'


    return dot_code


def generate_dot_network(configure_log, composite_algo_inst):
    # make connections
    dot_code ='\n// connecting the network\n'
    for algo_name, properties in configure_log.iteritems():
        for left_connector, right_connectors in properties['instance'].connections.iteritems():
            for right_connector in right_connectors:
                if isinstance(right_connector, streaming._StreamConnector):
                    dot_code += '        _'+inner_algo_name(left_connector.output_algo, configure_log).lower()+':'+left_connector.name+'_o:e'+' -> '+\
                                '_'+inner_algo_name(right_connector.input_algo, configure_log).lower()+':'+right_connector.name + '_i:w;\n'
                if isinstance(right_connector, types.NoneType):
                    inneralgoname = inner_algo_name(left_connector.output_algo, configure_log).lower()
                    dot_code += '        nowhere_'+inneralgoname+' [shape="box", style="rounded,filled", fillcolor="grey50", color="transparent" label="Nowhere" fontcolor="white" fontsize="18"];\n'+\
                                '        _'+inneralgoname+':'+left_connector.name+'_o:e'+' -> nowhere_'+inneralgoname+';\n'

    # make connections from floating inputs
    for name, connector in composite_algo_inst.inputs.iteritems():
        innerinputname = connector.name
        inneralgoname = inner_algo_name(connector.input_algo, configure_log).lower()
        dot_code += '        '+name+':e -> _'+inneralgoname+':'+innerinputname+'_i:w;\n'

    # make connections from floating outputs
    for name, connector in composite_algo_inst.outputs.iteritems():
        inneroutputname = connector.name
        inneralgoname = inner_algo_name(connector.output_algo, configure_log).lower()
        dot_code += '        _'+inneralgoname+':'+inneroutputname+'_o:e -> '+name+':w;\n'

    return dot_code


def generate_dot_cluster(configure_log, clustername, composite_algo_inst):
    ''' creates a cluster in dot language surrounded by a dashed, lightgrey line'''

    dot_code  = '    subgraph cluster_0 {\n'\
                '        color=lightgrey;\n'\
                '        style=dashed;\n'\
                '        label='+clustername+';\n\n'

    # for each algo in the cluster, declare it in dot:
    for algo_name, properties in configure_log.iteritems():
        dot_code += generate_dot_algo(algo_name, properties['instance'])

    # create the connections
    dot_code += generate_dot_network(configure_log, composite_algo_inst)

    # close the cluster code
    dot_code += '    }\n'

    return dot_code


def translate(composite_algo, output_filename, dot_graph=False):
    '''Takes in a class that is derived from essentia.streaming.CompositeBase and an output-filename
       and writes output-filename.h and output-filename.cpp versions of the given class.'''

    if not inspect.isclass(composite_algo):
        raise TypeError('"composite_algo" argument must be a class')

    if not streaming.CompositeBase in inspect.getmro(composite_algo):
        raise TypeError('"composite_algo" argument must inherit from essentia.streaming.CompositeBase')

    param_names, _, _, default_values = inspect.getargspec(composite_algo.__init__)
    param_names.remove('self')

    # these marker objects are used to track where config params travel in the network
    marker_objs = {}
    if not default_values and param_names: # python vars have no type so we cannot know what type they are!!
        raise TypeError('"composite_algo" arguments must have default values')
    if param_names:
        for param_name, value in zip(param_names, default_values):
            marker_objs[param_name] = MarkerObject(value)

    ### Before we call their function we need to neuter all of the configure methods of each
    ### streaming algorithm so that our markers won't cause the configure method to vomit

    configure_log = {}

    def dummy_configure(self, **kwargs):
        lbl = 0
        algo_name = self.name()+'_'+str(lbl)

        # increment lbl to generate a unique name for inner algo
        lowered_algo_names = [name.lower() for name in configure_log.keys()]
        while algo_name.lower() in lowered_algo_names:
            algo_name = algo_name[:algo_name.index('_')+1] + str(lbl)
            lbl +=1

        # algo_name is now unique

        configure_log[algo_name] = {}
        configure_log[algo_name]['instance'] = self
        configure_log[algo_name]['parameters'] = kwargs

        # We need to actually call the internal configure method because algorithms like silencerate
        # need to be configured so we can use its outputs. However we can't use our marker objects,
        # so we remove the marker objects that don't have a default value associated with them, and
        # for those that do have a default value, we use that value instead of the MarkerObject
        # itself
        kwargs_no_markers = dict(kwargs)

        for key, value in kwargs.iteritems():
            if value in marker_objs.values():
                if value.default_value == None:
                    del kwargs_no_markers[key]
                else:
                    kwargs_no_markers[key] = value.default_value

        self.real_configure(**kwargs_no_markers)

    # iterate over all streaming_algos
    streaming_algos = inspect.getmembers( streaming,
                                          lambda obj: inspect.isclass(obj) and \
                                                      _essentia.StreamingAlgorithm in inspect.getmro(obj) )
    streaming_algos = [member[1] for member in streaming_algos]
    for algo in streaming_algos:
        algo.real_configure = algo.configure
        algo.configure = dummy_configure

    ### Now generate an instance of their composite algorithm ###

    algo_inst = composite_algo(**marker_objs)

    # overwrite the dummy configure with the real configure method, so
    # translate can be called several times in the same file for a different
    # compositebase without entering in an infinite loop
    for algo in streaming_algos:
        algo.configure = algo.real_configure

    ### Do some checking on their network ###
    for algo in [ logitem['instance'] for logitem in configure_log.values() ]:
        if isinstance(algo, streaming.VectorInput):
            raise TypeError('essentia.streaming.VectorInput algorithms are not allowed for translatable composite algorithms')

        if isinstance(algo, streaming.AudioLoader) or \
           isinstance(algo, streaming.EasyLoader) or \
           isinstance(algo, streaming.MonoLoader) or \
           isinstance(algo, streaming.EqloudLoader):
            raise TypeError('No type of AudioLoader is allowed for translatable composite algorithms')

        if isinstance(algo, streaming.AudioWriter) or \
           isinstance(algo, streaming.MonoWriter):
            raise TypeError('No type of AudioWriter is allowed for translatable composite algorithms')

        if isinstance(algo, streaming.FileOutput):
            raise TypeError('essentia.streaming.FileOutput algorithms are not allowed for translatable composite algorithms')


    def sort_by_key(configure_log):
        # sort algorithms and conf values:
        sitems = configure_log.items()
        sitems.sort()
        sorted_algos = []
        sorted_params= []
        for k,v in sitems:
            sorted_params.append(v)
            sorted_algos.append(k)
        return sorted_algos, sorted_params

    sorted_algos, sorted_params = sort_by_key(configure_log)

    ### generate .h code ###

    h_code = '''// Generated automatically by essentia::translate

#ifndef STREAMING_''' + composite_algo.__name__.upper() + '''
#define STREAMING_''' + composite_algo.__name__.upper()+ '''

#include "streamingalgorithmcomposite.h"

class '''+composite_algo.__name__+''' : public essentia::streaming::AlgorithmComposite {
 protected:
'''

    for algo_name in sorted_algos:
        h_code += '  essentia::streaming::Algorithm* _'+algo_name.lower()+';\n'

    h_code += '''
 public:
  '''+composite_algo.__name__+'''();

  ~'''+composite_algo.__name__+'''() {
'''
    for algo_name in sorted_algos:
        h_code += '    delete _'+algo_name.lower()+';\n'

    h_code += '''  }

  void declareParameters() {
'''
    if param_names:
        for param_name, default_value in zip(param_names, default_values):
            h_code += '    declareParameter("'+param_name+'", "", "", '

            if isinstance(default_value, basestring): h_code += '"'+default_value+'"'
            else:                                     h_code += str(default_value)

            h_code += ');\n'

    h_code += '''  }

  void configure();
  void createInnerNetwork();
  void reset();

  static const char* name;
  static const char* version;
  static const char* description;
};
#endif
'''

    ### Generate .cpp code ###

    cpp_code = '''// Generated automatically by essentia::translate

#include "'''+output_filename+'''.h"
#include "algorithmfactory.h"
#include "taskqueue.h"

using namespace std;
using namespace essentia;
using namespace essentia::streaming;

const char* '''+composite_algo.__name__+'''::name = "'''+composite_algo.__name__+'''";
const char* '''+composite_algo.__name__+'''::version = "1.0";
const char* '''+composite_algo.__name__+'''::description = DOC("");\n\n'''


################################
# CONSTRUCTOR
################################
    cpp_code += composite_algo.__name__+'''::'''+composite_algo.__name__+'''(): '''
    for algo_name in sorted_algos: cpp_code += '_' + algo_name.lower() + '(0), '
    cpp_code = cpp_code[:-2] + ''' {
  setName("''' + composite_algo.__name__ + '''");
  declareParameters();
  AlgorithmFactory& factory = AlgorithmFactory::instance();\n\n'''

    # create inner algorithms
    for algo_name in sorted_algos:
        cpp_code += '  _'+algo_name.lower()+' = factory.create("'+algo_name[:algo_name.rindex('_')]+'");\n'

    cpp_code+='}\n\n'

################################
# INNER NETWORK
################################
# declaration of inputs and output and connecting the network should not be
# done in the constructor, as there are algos like silencerate which
# inputs/outputs depend on the configuration parameters. Hence, it is safer to
# do it in the configure() function

    cpp_code += 'void ' + composite_algo.__name__ + '::createInnerNetwork() {\n'

    # declare inputs
    for input_alias, connector in algo_inst.inputs.iteritems():
        input_owner_name = None
        input_name = None

        for algo_name, properties in zip(sorted_algos, sorted_params): #configure_log.iteritems():
            if properties['instance'] == connector.input_algo:
                input_owner_name = algo_name
                input_name = connector.name
                break

        if not input_owner_name:
            raise RuntimeError('Could not determine owner of the \''+input_alias+'\' input')

        cpp_code += '  declareInput(_'+input_owner_name.lower()+'->input("'+input_name+'"), "'+input_alias+'", "");\n'

    cpp_code += '\n'

    # declare outputs
    aliases, connectors = sort_by_key(algo_inst.outputs)
    for output_alias, connector in zip(aliases, connectors):
        output_owner_name = None
        output_name = None

        for algo_name, properties in zip(sorted_algos, sorted_params): #configure_log.iteritems():
            if properties['instance'] == connector.output_algo:
                output_owner_name = algo_name
                output_name = connector.name
                break

        if not output_owner_name:
            raise RuntimeError('Could not determine owner of the \''+output_alias+'\' output')

        cpp_code += '  declareOutput(_'+output_owner_name.lower()+'->output("'+output_name+'"), "'+output_alias+'", "");\n'

    cpp_code += '\n'

    # make connections
    for algo_name, properties in zip(sorted_algos, sorted_params): #configure_log.iteritems():
        for left_connector, right_connectors in properties['instance'].connections.iteritems():
            for right_connector in right_connectors:
                if isinstance(right_connector, streaming._StreamConnector):
                    cpp_code += '  connect( _'+\
                                inner_algo_name(left_connector.output_algo, configure_log).lower() + \
                                '->output("'+left_connector.name+'"), _' + \
                                inner_algo_name(right_connector.input_algo, configure_log).lower() + \
                                '->input("'+right_connector.name+'") );\n'

                elif isinstance(right_connector, types.NoneType):
                    cpp_code += '  connect( _'+\
                                inner_algo_name(left_connector.output_algo, configure_log).lower() + \
                                '->output("'+left_connector.name+'"), NOWHERE );\n'

    cpp_code = cpp_code[:-1]
    cpp_code += '''
}\n\n'''


################################
# CONFIGURE
################################

    cpp_code += 'void '+composite_algo.__name__+'::configure() {\n'

    # configure method

    # create local variable for every composite parameter
    for composite_param_name in param_names:
        param_edt = find_edt(composite_param_name, marker_objs[composite_param_name], configure_log)
        cpp_code += '  '+edt_cpp_code[param_edt]+' '+composite_param_name + \
                    ' = parameter("'+composite_param_name+'").to' + \
                    edt_parameter_code[param_edt]+'();\n'

    cpp_code += '\n'

    # configure inner algorithms
    for algo_name, properties in zip(sorted_algos, sorted_params): #configure_log.iteritems():
        # skip if inner algorithm wasn't configured explicitly
        if not properties['parameters']: continue

        for param_name, value in properties['parameters'].iteritems():
            type = common.determineEdt(value)
            if 'LIST' in str(type) or 'VECTOR' in str(type):
                if type in [common.Edt.VECTOR_STRING]:
                    cpp_code += '  const char* ' + param_name + '[]  = {'
                    for s in value: cpp_code += '\"' + s + '\"' + ','
                elif type in[common.Edt.VECTOR_REAL, common.Edt.LIST_REAL]:
                    cpp_code += '  Real ' + param_name + '[] = {'
                    for f in value: cpp_code +=  str(f) + ','
                elif type in [common.Edt.VECTOR_INT, common.Edt.LIST_INT]:
                    cpp_code += '  int' + param_name + '[] = {'
                    for i in value: cpp_code +=  str(i) + ','
                cpp_code = cpp_code[:-1]+'};\n'

        cpp_code += '  _'+algo_name.lower()+'->configure('

        for param_name, value in properties['parameters'].iteritems():
            if isinstance(value, MarkerObject):
                # figure out which composite param it is
                composite_param_name = None
                for marker_name, marker_obj in marker_objs.iteritems():
                    if marker_obj == value:
                        composite_param_name = marker_name
                        break

                if not composite_param_name:
                    raise RuntimeError('Could not determine which composite parameter to use to configure inner algorithm \''+algo_name+'\'s parameter \''+param_name+'\'')

                cpp_code += '"'+param_name+'", '+composite_param_name+', '

            else:
                type = common.determineEdt(value)
                if 'LIST' in str(type) or 'VECTOR' in str(type):
                    if type in [common.Edt.VECTOR_STRING]:
                        cpp_code += '"'+param_name+'", '+'arrayToVector<string>(' + param_name + ')  '
                    elif type in[common.Edt.VECTOR_REAL, common.Edt.LIST_REAL]:
                        cpp_code += '"'+param_name+'", '+'arrayToVector<Real>(' + param_name + ')  '
                    elif type in [common.Edt.VECTOR_INT, common.Edt.LIST_INT]:
                        cpp_code += '"'+param_name+'", '+'arrayToVector<int>(' + param_name + ')  '
                elif isinstance(value, basestring):
                    cpp_code += '"'+param_name+'", "'+value+'", '
                elif isinstance(value, bool):
                    if value: cpp_code += '"'+param_name+'", true, '
                    else:     cpp_code += '"'+param_name+'", false, '
                else:
                    cpp_code += '"'+param_name+'", '+str(value)+', '

        cpp_code = cpp_code[:-2] + ');\n'
    cpp_code += '  createInnerNetwork();\n}\n\n'

################################
# RESET
################################

    cpp_code += 'void '+composite_algo.__name__+'::reset() {\n'
    for algo_name in sorted_algos:
        cpp_code += '  _' + algo_name.lower() + '->reset();\n'
    cpp_code += '}\n\n'

################################
# DESTRUCTOR
################################
# see h_code. Each algo from the composite is deleted separately instead of
# calling deleteNetwork
#    cpp_code += composite_algo.__name__+'''::~'''+composite_algo.__name__+'''() {
#  deleteNetwork(_''' + input_owner_name.lower() + ''');
#}'''
#    cpp_code +='\n'
#
################################
# end of cpp code
################################

    if dot_graph:
        ### generate .dot code ###
        dot_code  = 'digraph ' + output_filename +' {\n'
        dot_code += '    rankdir=LR\n' # if instead of top-down left-right is prefered
        # general formatting options:
        dot_code += '    node [color=black, fontname=Verdana, weight=1, fontsize=8, shape=Mrecord]\n'
        dot_code += '    edge [color=black, style=solid, weight=1, arrowhead="dotnormal", arrowtail="dot", arrowsize=1, fontsize=6]\n'

        # for each input generate nodes
        for name in algo_inst.inputs.keys():
            dot_code += '    '+name+' [label="'+name+'"];\n'

        dot_code += generate_dot_cluster(configure_log, composite_algo.__name__, algo_inst)

        # for each output generate nodes
        for name in algo_inst.outputs.keys():
            dot_code += '    '+name+' [label="'+name+'"];\n'

        dot_code += '}'

    ### Write files ###

    f = open(output_filename+'.h', 'w')
    f.write(h_code)
    f.close()

    f = open(output_filename+'.cpp', 'w')
    f.write(cpp_code)
    f.close()

    if dot_graph:
        f = open(output_filename+'.dot', 'w')
        f.write(dot_code)
        f.close()




# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from argparse import ArgumentParser
from multiprocessing import Pool
from multiprocessing import cpu_count
from subprocess import Popen, PIPE
from essentia import EssentiaError
import os
import sys


def __subprocess__(cmd):
    """
    General purpose subprocess.
    """

    cmd_str = ' '.join(cmd)
    print('Running "{}"...'.format(cmd_str))

    child = Popen(cmd, stdout=None, stderr=PIPE)
    _, stderr = child.communicate()
    rc = child.returncode

    return rc, cmd_str, stderr.decode("utf-8")


def batch_music_extractor(audio_dir, output_dir, generate_log=True,
                          audio_types=None, profile=None,
                          store_frames=False, skip_analyzed=False,
                          format='yaml', jobs=0):
    """
        Processes every audio file matching `audio_types` in `audio_dir` with MusicExtractor. The generated
        .sig yaml/json files are stored in `output_dir` matching the folder structure found in `audio_dir`.
    """

    if not audio_types:
        audio_types = ('.WAV', '.AIFF', '.FLAC', '.MP3', '.OGG')
        print("Audio files extensions considered by default: " +
              ' '.join(audio_types))
    else:
        audio_types = tuple(audio_types)
        print("Searching for audio files extensions: " + ' '.join(audio_types))
    print("")

    output_dir = os.path.abspath(output_dir)

    if profile:
        assert os.path.isfile(profile)

    if jobs == 0:
        try:
            jobs = cpu_count()
        except NotImplementedError:
            print("Failed to automatically detect the cpu count, the analysis will try to continue with 4 jobs. For a different behavior change the `job` parameter.")
            jobs = 4

    # find all audio files and prepare folder structure in the output folder
    os.chdir(audio_dir)

    skipped_count = 0
    skipped_files = []
    cmd_lines = []
    for root, dirnames, filenames in os.walk("."):
        for filename in filenames:
            if filename.upper().endswith(audio_types):
                audio_file = os.path.relpath(os.path.join(root, filename))
                audio_file_abs = os.path.join(audio_dir, audio_file)
                sig_file = os.path.join(output_dir, audio_file)

                if skip_analyzed:
                    if os.path.isfile(sig_file + '.sig'):
                        print("Found descriptor file for " +
                              audio_file + ", skipping...")
                        skipped_files.append(audio_file)
                        skipped_count += 1
                        continue
                folder = os.path.dirname(sig_file)
                if not os.path.exists(folder):
                    os.makedirs(folder)

                elif os.path.isfile(folder):
                    raise EssentiaError('Cannot create directory {} .There exist a file with the same name. Aborting analysis.'.format(folder))

                cmd_line = [
                            sys.executable,
                            os.path.join(os.path.dirname(__file__), 'extractors/music_extractor.py'), 
                            audio_file_abs, sig_file,
                            '--format', format
                            ]

                if store_frames:
                    cmd_line += ['--store_frames']

                if profile:
                    cmd_line += ['--profile', profile]

                cmd_lines.append(cmd_line)


    # analyze
    errors, oks = 0, 0
    if len(cmd_lines) > 0:
        p = Pool(jobs)
        outs = p.map(__subprocess__, cmd_lines)

        status, cmd, stderr = zip(*outs)

        oks, errors = 0, 0
        for i in status:
            if i == 0:
                oks += 1
            else:
                errors += 1

    summary = "Analysis done. {} files have been skipped due to errors, {} were processed and {} already existed.".format(
        errors, oks, skipped_count)
    print(summary)

    # generate log
    if generate_log:
        log = [summary]

        if errors > 0:
            log += ['Errors:'] + ['"{}"\n{}\n\n'.format(cmd[idx], stderr[idx])
                                  for idx, i in enumerate(status) if i != 0]

        if oks > 0:
            log += ['Oks:'] + [cmd[idx]
                               for idx, i in enumerate(status) if i == 0]

        if skipped_count > 0:
            log += ['Skipped files:'] + skipped_files

        if not os.path.exists(output_dir):
            os.makedirs(output_dir)

        with open(os.path.join(output_dir, 'log'), 'w') as f:
            f.write('\n'.join(log))



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

def create_python_algorithms(essentia):

    '''
    # Spectral Decrease
    class SpectralDecrease(essentia.Decrease):
        def configure(self, sampleRate = 44100, **kwargs):
            essentia.Decrease.configure(self, range = sampleRate*0.5)

    setattr(essentia, 'SpectralDecrease', SpectralDecrease)


    # AudioDecrease
    class AudioDecrease(essentia.Decrease):
        def configure(self, blockSize, sampleRate = 44100, **kwargs):
            essentia.Decrease.configure(self, range = (blockSize-1.0)/sampleRate)

    setattr(essentia, 'AudioDecrease', AudioDecrease)


    # SpectralCentroid
    class SpectralCentroid(essentia.Centroid):
        def configure(self, sampleRate = 44100, **kwargs):
            essentia.Centroid.configure(self, range = sampleRate*0.5)

    setattr(essentia, 'SpectralCentroid', SpectralCentroid)


    # AudioCentroid
    class AudioCentroid(essentia.Centroid):
        def configure(self, blockSize, sampleRate = 44100, **kwargs):
            essentia.Centroid.configure(self, range = (blockSize-1.0)/sampleRate)

    setattr(essentia, 'AudioCentroid', AudioCentroid)


    # SpectralCentralMoments
    class SpectralCentralMoments(essentia.CentralMoments):
        def configure(self, sampleRate = 44100, **kwargs):
            essentia.CentralMoments.configure(self, range = sampleRate*0.5)

    setattr(essentia, 'SpectralCentralMoments', SpectralCentralMoments)


    # AudioCentralMoments
    class AudioCentralMoments(essentia.CentralMoments):
        def configure(self, blockSize, sampleRate = 44100, **kwargs):
            essentia.CentralMoments.configure(self, range = (blockSize-1.0)/sampleRate)

    setattr(essentia, 'AudioCentralMoments', AudioCentralMoments)
    '''

    default_fc = essentia.FrameCutter()



    # FrameGenerator
    class FrameGenerator(object):
        __struct__ = { 'name': 'FrameGenerator',
                       'category': 'Standard',
                       'inputs': [],
                       'outputs': [],
                       'parameters': default_fc.__struct__['parameters'],
                       'description': '''The FrameGenerator is a Python generator for the FrameCutter algorithm. It is not available in C++.

FrameGenerator inherits all the parameters of the FrameCutter. The way to use it in Python is the following:

  for frame in FrameGenerator(audio, frameSize, hopSize):
      do_something()

''' }


        def __init__(self, audio, frameSize = default_fc.paramValue('frameSize'), 
                                  hopSize = default_fc.paramValue('hopSize'),
                                  startFromZero = default_fc.paramValue('startFromZero'),
                                  validFrameThresholdRatio = default_fc.paramValue('validFrameThresholdRatio'),
                                  lastFrameToEndOfFile=default_fc.paramValue('lastFrameToEndOfFile')):

            self.audio = audio
            self.frameSize = frameSize
            self.hopSize = hopSize
            self.startFromZero = startFromZero
            self.validFrameThresholdRatio = validFrameThresholdRatio
            self.lastFrameToEndOfFile=lastFrameToEndOfFile
            self.frame_creator = essentia.FrameCutter(frameSize = frameSize,
                                                       hopSize = hopSize,
                                                       startFromZero = startFromZero,
                                                       validFrameThresholdRatio=validFrameThresholdRatio,
                                                       lastFrameToEndOfFile=lastFrameToEndOfFile)

        def __iter__(self):
            return self

        def __next__(self):
            frame = self.frame_creator.compute(self.audio)
            if frame.size == 0:
                raise StopIteration
            else:
                return frame

        next = __next__  # Python 2

        def num_frames(self):
            if self.startFromZero:
                if not self.lastFrameToEndOfFile:
                    size = int(round(len(self.audio) / float(self.hopSize)))
                else:
                    size = int(round((len(self.audio)+self.frameSize) / float(self.hopSize)))
            else:
                size = int(round((len(self.audio)+self.frameSize/2.0)/float(self.hopSize)))
            return size

        def frame_times(self, sampleRate):
            times = []
            if self.startFromZero:
                start = self.frameSize/2.0
            else:
                start = 0.

            for i in range(self.num_frames()):
                times.append((start + self.hopSize * i) / sampleRate)
            return times


    setattr(essentia, 'FrameGenerator', FrameGenerator)



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from essentia.standard import MusicExtractor, YamlOutput
from essentia import EssentiaError
from argparse import ArgumentParser
import os
import sys


def music_extractor(audio_file, sig_file, profile=None, store_frames=False, format='yaml'):
    if profile:
        extractor = MusicExtractor(profile=profile)
    else:
        extractor = MusicExtractor()

    poolStats, poolFrames = extractor(audio_file)

    folder = os.path.dirname(sig_file)

    if not os.path.exists(folder):
        os.makedirs(folder)
    elif os.path.isfile(folder):
        raise EssentiaError('Cannot create directory {} .There exist a file with the same name. Aborting analysis.'.format(folder))

    output = YamlOutput(filename=sig_file+'.sig', format=format)
    output(poolStats)
    if store_frames:
        YamlOutput(filename=sig_file + '.frames.sig', format=format)(poolFrames)


if __name__ == '__main__':
    parser = ArgumentParser(description = """
Analyzes an audio file using MusicExtractor.
""")

    parser.add_argument('audio_file', help='audio file name')
    parser.add_argument('sig_file', help='sig file name')
    parser.add_argument('--profile', help='MusicExtractor profile', required=False)
    parser.add_argument('--store_frames', help='store frames data', action='store_true', required=False)
    parser.add_argument('--format', help='yaml or json', default='yaml', choices=['yaml', 'json'])
    args = parser.parse_args()

    music_extractor(args.audio_file, args.sig_file, profile=args.profile, store_frames=args.store_frames, format=args.format)



#!/usr/bin/env python

import math
import sys
import os
import random
import struct
import popen2
import getopt
import numpy

pi=math.pi
e=math.e
j=complex(0,1)

doreal=0

datatype = os.environ.get('DATATYPE','float')

util = '../tools/fft_' + datatype
minsnr=90
if datatype == 'double':
    fmt='d'
elif datatype=='int16_t':
    fmt='h'
    minsnr=10
elif datatype=='int32_t':
    fmt='i'
elif datatype=='simd':
    fmt='4f'
    sys.stderr.write('testkiss.py does not yet test simd')
    sys.exit(0)
elif datatype=='float':
    fmt='f'
else:
    sys.stderr.write('unrecognized datatype %s\n' % datatype)
    sys.exit(1)
 

def dopack(x,cpx=1):
    x = numpy.reshape( x, ( numpy.size(x),) )
    
    if cpx:
        s = ''.join( [ struct.pack(fmt*2,c.real,c.imag) for c in x ] )
    else:
        s = ''.join( [ struct.pack(fmt,c.real) for c in x ] )
    return s

def dounpack(x,cpx):
    uf = fmt * ( len(x) / struct.calcsize(fmt) )
    s = struct.unpack(uf,x)
    if cpx:
        return numpy.array(s[::2]) + numpy.array( s[1::2] )*j
    else:
        return numpy.array(s )

def make_random(dims=[1]):
    res = []
    for i in range(dims[0]):
        if len(dims)==1:
            r=random.uniform(-1,1)
            if doreal:
                res.append( r )
            else:
                i=random.uniform(-1,1)
                res.append( complex(r,i) )
        else:
            res.append( make_random( dims[1:] ) )
    return numpy.array(res)

def flatten(x):
    ntotal = numpy.size(x)
    return numpy.reshape(x,(ntotal,))

def randmat( ndims ):
    dims=[]
    for i in range( ndims ):
        curdim = int( random.uniform(2,5) )
        if doreal and i==(ndims-1):
            curdim = int(curdim/2)*2 # force even last dimension if real
        dims.append( curdim )
    return make_random(dims )

def test_fft(ndims):
    x=randmat( ndims )


    if doreal:
        xver = numpy.fft.rfftn(x)
    else:
        xver = numpy.fft.fftn(x)
    
    open('/tmp/fftexp.dat','w').write(dopack( flatten(xver) , True ) )

    x2=dofft(x,doreal)
    err = xver - x2
    errf = flatten(err)
    xverf = flatten(xver)
    errpow = numpy.vdot(errf,errf)+1e-10
    sigpow = numpy.vdot(xverf,xverf)+1e-10
    snr = 10*math.log10(abs(sigpow/errpow) )
    print 'SNR (compared to NumPy) : %.1fdB' % float(snr)

    if snr<minsnr:
        print 'xver=',xver
        print 'x2=',x2
        print 'err',err
        sys.exit(1)
 
def dofft(x,isreal):
    dims=list( numpy.shape(x) )
    x = flatten(x)

    scale=1
    if datatype=='int16_t':
        x = 32767 * x
        scale = len(x) / 32767.0
    elif datatype=='int32_t':
        x = 2147483647.0 * x
        scale = len(x) / 2147483647.0

    cmd='%s -n ' % util
    cmd += ','.join([str(d) for d in dims])
    if doreal:
        cmd += ' -R '

    print cmd
    p = popen2.Popen3(cmd )

    open('/tmp/fftin.dat','w').write(dopack( x , isreal==False ) )

    p.tochild.write( dopack( x , isreal==False ) )
    p.tochild.close()

    res = dounpack( p.fromchild.read() , 1 )
    open('/tmp/fftout.dat','w').write(dopack( flatten(res) , True ) )
    if doreal:
        dims[-1] = int( dims[-1]/2 ) + 1

    res = scale * res

    p.wait()
    return numpy.reshape(res,dims)

def main():
    opts,args = getopt.getopt(sys.argv[1:],'r')
    opts=dict(opts)

    global doreal
    doreal = opts.has_key('-r')

    if doreal:
        print 'Testing multi-dimensional real FFTs'
    else:
        print 'Testing multi-dimensional FFTs'

    for dim in range(1,4):
        test_fft( dim )

if __name__ == "__main__":
    main()




# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from six import iteritems
from . import _essentia
import essentia
from . import common as _c
import sys as _sys
from ._essentia import keys as algorithmNames, info as algorithmInfo
from copy import copy

# given an essentia algorithm name, create the corresponding class
def _create_essentia_class(name, moduleName = __name__):
    essentia.log.debug(essentia.EPython, 'Creating essentia.standard class: %s' % name)

    _algoInstance = _essentia.Algorithm(name)
    _algoDoc = _algoInstance.getDoc()
    _algoStruct = _algoInstance.getStruct()
    del _algoInstance

    class Algo(_essentia.Algorithm):
        __doc__ = _algoDoc
        __struct__ = _algoStruct

        def __init__(self, **kwargs):
            # init the internal cpp wrapper
            _essentia.Algorithm.__init__(self, name)

            # configure the algorithm
            self.configure(**kwargs)

        def configure(self, **kwargs):
            # verify that all types match and do any necessary conversions
            for name, val in iteritems(kwargs):
                goalType = self.paramType(name)

                if type(val).__module__ == 'numpy':
                    if not val.flags['C_CONTIGUOUS']:
                        val = copy(val)

                try:
                    convertedVal = _c.convertData(val, goalType)
                except TypeError: # as e: # catching exception as sth is only
                                          #available as from python 2.6
                    raise TypeError('Error cannot convert parameter %s to %s'\
                                    %(str(_c.determineEdt(val)),str(goalType))) #\''+name+'\' parameter: '+str(e))

                kwargs[name] = convertedVal

            self.__configure__(**kwargs)

        def compute(self, *args):
            inputNames = self.inputNames()

            if len(args) != len(inputNames):
                raise ValueError(name+'.compute requires '+str(len(inputNames))+' argument(s), '+str(len(args))+' given')

            # we have to make some exceptions for YamlOutput and PoolAggregator
            # because they expect cpp Pools
            if name in ('YamlOutput', 'PoolAggregator', 'SvmClassifier', 'PCA', 'GaiaTransform', 'TensorflowPredict'):
                args = (args[0].cppPool,)

            # verify that all types match and do any necessary conversions
            result = []

            convertedArgs = []

            for i in range(len(inputNames)):
                arg = args[i]

                if type(args[i]).__module__ == 'numpy':
                    if arg.dtype == 'float64':
                        arg = arg.astype('float32')
                        essentia.INFO('Warning: essentia can currently only accept numpy arrays of dtype '
                                      '"single". "%s" dtype is double. Precision will be automatically '
                                      'truncated into "single".' %(inputNames[i]))
                    if not args[i].flags['C_CONTIGUOUS']:
                        arg = copy(args[i])

                goalType = _c.Edt(self.inputType(inputNames[i]))

                try:
                    convertedData = _c.convertData(arg, goalType)
                except TypeError:
                    raise TypeError('Error cannot convert argument %s to %s' \
                          %(str(_c.determineEdt(arg)), str(goalType)))

                convertedArgs.append(convertedData)

            results = self.__compute__(*convertedArgs)

            # we have to make an exceptional case for YamlInput, because we need
            # to wrap the Pool that it outputs w/ our python Pool from common.py
            if name in ('YamlInput', 'PoolAggregator', 'SvmClassifier', 'PCA', 'GaiaTransform', 'Extractor', 'TensorflowPredict'):
                return _c.Pool(results)

            # MusicExtractor and FreesoundExtractor output two pools
            if name in ('MusicExtractor', 'FreesoundExtractor'):
                return (_c.Pool(results[0]), _c.Pool(results[1]))

            # In the case of MetadataReader, the 7th output is also a Pool
            if name in ('MetadataReader'):
                return results[:7] + (_c.Pool(results[7]),) + results[8:]

            else:
                return results

        def __call__(self, *args):
            return self.compute(*args)

        def __str__(self):
            return __doc__


    algoClass = _c.algoDecorator(Algo)

    setattr(_sys.modules[moduleName], name, algoClass)

    return algoClass


# load all classes into python
def _reloadAlgorithms(moduleName = __name__):
    for name in _essentia.keys():
        _create_essentia_class(name, moduleName)

_reloadAlgorithms()


# load derived descriptors and other ones written in python
from .algorithms import create_python_algorithms as _create_python_algorithms
_create_python_algorithms(_sys.modules[__name__])



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

from . import common as _c
from . import _essentia

def isSilent(arg):
    return _essentia.isSilent( _c.convertData(arg, _c.Edt.VECTOR_REAL) )

def instantPower(arg):
    return _essentia.instantPower( _c.convertData(arg, _c.Edt.VECTOR_REAL) )

def nextPowerTwo(arg):
    return _essentia.nextPowerTwo(_c.convertData(arg, _c.Edt.REAL))

def isPowerTwo(arg):
    return _essentia.isPowerTwo(arg)

def lin2db(arg):
    return _essentia.lin2db( _c.convertData(arg, _c.Edt.REAL) )

def db2lin(arg):
    return _essentia.db2lin( _c.convertData(arg, _c.Edt.REAL) )

def pow2db(arg):
    return _essentia.pow2db( _c.convertData(arg, _c.Edt.REAL) )

def db2pow(arg):
    return _essentia.db2pow( _c.convertData(arg, _c.Edt.REAL) )

def amp2db(arg):
    return _essentia.amp2db( _c.convertData(arg, _c.Edt.REAL) )

def db2amp(arg):
    return _essentia.db2amp( _c.convertData(arg, _c.Edt.REAL) )

def lin2log(arg):
    return _essentia.lin2log( _c.convertData(arg, _c.Edt.REAL) )

def bark2hz(arg):
    return _essentia.bark2hz( _c.convertData(arg, _c.Edt.REAL) )

def hz2bark(arg):
    return _essentia.hz2bark( _c.convertData(arg, _c.Edt.REAL) )

def mel2hz(arg):
    return _essentia.mel2hz( _c.convertData(arg, _c.Edt.REAL) )

def hz2mel(arg):
    return _essentia.hz2mel( _c.convertData(arg, _c.Edt.REAL) )

def equivalentKey(arg):
    return _essentia.equivalentKey( _c.convertData(arg, _c.Edt.STRING) )

def postProcessTicks(arg1, arg2=None, arg3=None):
    if arg2 != None and arg3 != None:
        return _essentia.postProcessTicks(_c.convertData(arg1, _c.Edt.VECTOR_REAL),
                                          _c.convertData(arg2, _c.Edt.VECTOR_REAL),
                                          _c.convertData(arg3, _c.Edt.REAL))
    return _essentia.postProcessTicks(_c.convertData(arg1, _c.Edt.VECTOR_REAL))

def normalize(array):
    return _essentia.normalize(_c.convertData(array, _c.Edt.VECTOR_REAL))

def derivative(array):
    return _essentia.derivative(_c.convertData(array, _c.Edt.VECTOR_REAL))

__all__ = [ 'isSilent', 'instantPower',
            'nextPowerTwo', 'isPowerTwo',
            'lin2db', 'db2lin',
            'pow2db', 'db2pow',
            'amp2db', 'db2amp',
            'bark2hz', 'hz2bark',
            'mel2hz', 'hz2mel',
            'postProcessTicks',
            'normalize', 'derivative',
            'equivalentKey', 'lin2log']



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

import essentia.standard
import tempfile
import numpy as np

def test_audiofile(filename=None, type='sin440', duration=1):

    """Create a dummy wav audio file.
    This can be useful for tests that require loading audio files.
    
    Args:
        filename (string): Filename (default=None). If not specified, a 
            named temporary file will be created. 
        type (string): The type of audio to generate: 'silence' or 'sin440' 
            (default)
        duration (float): duration of audio in seconds (default 1 sec.)
    Returns:
        (string): Name of a temporary audio file
    """
    if type == 'sin440':
        samples = np.sin(2*np.pi*np.arange(44100*duration)*440/44100).astype(np.float32)
    elif type == 'silence':
        samples = [0] * int(44100 * duration)
    else:
        raise (ValueError, 'Wrong audio type:', type)
    
    if not filename:
        tf = tempfile.NamedTemporaryFile(delete=False, suffix=".wav")
        tf.close()
        filename = tf.name

    essentia.standard.MonoWriter(filename=filename)(samples)
    return filename



#!/usr/bin/env python

import FFT
import sys
import random
import re
j=complex(0,1)

def randvec(n,iscomplex):
    if iscomplex:
        return [
                int(random.uniform(-32768,32767) ) + j*int(random.uniform(-32768,32767) )
                for i in range(n) ]
    else:                
        return [ int(random.uniform(-32768,32767) ) for i in range(n) ]
    
def c_format(v,round=0):
    if round:
        return ','.join( [ '{%d,%d}' %(int(c.real),int(c.imag) ) for c in v ] ) 
    else:
        s= ','.join( [ '{%.60f ,%.60f }' %(c.real,c.imag) for c in v ] ) 
        return re.sub(r'\.?0+ ',' ',s)

def test_cpx( n,inverse ,short):
    v = randvec(n,1)
    scale = 1
    if short:
        minsnr=30
    else:
        minsnr=100

    if inverse:
        tvecout = FFT.inverse_fft(v)
        if short:
            scale = 1
        else:            
            scale = len(v)
    else:
        tvecout = FFT.fft(v)
        if short:
            scale = 1.0/len(v)

    tvecout = [ c * scale for c in tvecout ]


    s="""#define NFFT %d""" % len(v) + """
    {
        double snr;
        kiss_fft_cpx test_vec_in[NFFT] = { """  + c_format(v) + """};
        kiss_fft_cpx test_vec_out[NFFT] = {"""  + c_format( tvecout ) + """};
        kiss_fft_cpx testbuf[NFFT];
        void * cfg = kiss_fft_alloc(NFFT,%d,0,0);""" % inverse + """

        kiss_fft(cfg,test_vec_in,testbuf);
        snr = snr_compare(test_vec_out,testbuf,NFFT);
        printf("DATATYPE=" xstr(kiss_fft_scalar) ", FFT n=%d, inverse=%d, snr = %g dB\\n",NFFT,""" + str(inverse) + """,snr);
        if (snr<""" + str(minsnr) + """)
            exit_code++;
        free(cfg);
    }
#undef NFFT    
"""
    return s

def compare_func():
    s="""
#define xstr(s) str(s)
#define str(s) #s
double snr_compare( kiss_fft_cpx * test_vec_out,kiss_fft_cpx * testbuf, int n)
{
    int k;
    double sigpow,noisepow,err,snr,scale=0;
    kiss_fft_cpx err;
    sigpow = noisepow = .000000000000000000000000000001; 

    for (k=0;k<n;++k) {
        sigpow += test_vec_out[k].r * test_vec_out[k].r + 
                  test_vec_out[k].i * test_vec_out[k].i;
        C_SUB(err,test_vec_out[k],testbuf[k].r);
        noisepow += err.r * err.r + err.i + err.i;

        if (test_vec_out[k].r)
            scale += testbuf[k].r / test_vec_out[k].r;
    }
    snr = 10*log10( sigpow / noisepow );
    scale /= n;
    if (snr<10)
        printf( "\\npoor snr, try a scaling factor %f\\n" , scale );
    return snr;
}
"""
    return s

def main():

    from getopt import getopt
    opts,args = getopt(sys.argv[1:],'s')
    opts = dict(opts)
    short = int( opts.has_key('-s') )

    fftsizes = args
    if not fftsizes:
        fftsizes = [ 1800 ]
    print '#include "kiss_fft.h"'
    print compare_func()
    print "int main() { int exit_code=0;\n"
    for n in fftsizes:
        n = int(n)
        print test_cpx(n,0,short)
        print test_cpx(n,1,short)
    print """
    return exit_code;
}
"""

if __name__ == "__main__":
    main()



from . import _essentia
import sys as _sys
from ._essentia import reset
from .common import Pool, array, ones, zeros
from .progress import Progress
from .utils import *


__version__ = _essentia.version()
__version_git_sha__ = _essentia.version_git_sha();


# base Exception class
class EssentiaError(Exception):
    def __init__(self, error, filename=None):
        self.error = error
        self.filename = filename
    def __str__(self):
        if self.filename:
            return "ERROR: " + self.error.replace(';',':') + self.filename
        else:
            return "ERROR: " + self.error.replace(';',':')


# set up logging system

# debug level messages (separate handling because they are split into modules)
class DebuggingModule:
    EAlgorithm   = 1 << 0
    EConnectors  = 1 << 1
    EFactory     = 1 << 2
    ENetwork     = 1 << 3
    EGraph       = 1 << 4
    EExecution   = 1 << 5
    EMemory      = 1 << 6
    EScheduler   = 1 << 7

    EPython      = 1 << 20
    EPyBindings  = 1 << 21
    EUnittest    = 1 << 22

    EUser1       = 1 << 25
    EUser2       = 1 << 26

    ENone        = 0
    EAll         = (1 << 30) - 1

# import the DebuggingModule enum directly in our namespace for convenience
for dbg in dir(DebuggingModule):
    if not dbg.startswith('E'): continue
    setattr(_sys.modules[__name__], dbg, getattr(DebuggingModule, dbg))


class BitMask(object):
    def __init__(self, value = 0):
        self._value = value

    def __or__(self, other):
        self._value |= other

    def __add__(self, other):
        return self._value | other

    def __iadd__(self, other):
        self._value |= other
        return self

    def __and__(self, other):
        self._value &= other

    def __invert__(self):
        self._value = ~self._value

    def __sub__(self, other):
        return self._value & ~other

    def __isub__(self, other):
        self._value &= ~other
        return self

    def __int__(self):
        return self._value

    def __repr__(self):
        active, inactive = [], []
        for dbg in dir(DebuggingModule):
            if (not dbg.startswith('E') or
                dbg == 'EAll' or dbg == 'ENone'):
                continue
            if self._value & getattr(DebuggingModule, dbg):
                active.append(dbg[1:])
            else:
                inactive.append(dbg[1:])
        return 'DebugLevels(On=[%s], Off=[%s])' % (','.join(active), ','.join(inactive))



def _getDebugLevel(_):
    return BitMask(_essentia.debugLevel())

def _setDebugLevel(_, levels):
    _essentia.setDebugLevel(int(levels))


class EssentiaLogger(object):
    debugLevels = property(_getDebugLevel, _setDebugLevel,
                           doc = 'A bit mask indicating which debugging modules should be activated')

    infoActive = property(lambda l: _essentia.infoLevel(), lambda l, active: _essentia.setInfoLevel(active),
                          doc = 'A boolean indicating whether messages at the info level should be displayed')

    warningActive = property(lambda l: _essentia.warningLevel(), lambda l, active: _essentia.setWarningLevel(active),
                             doc = 'A boolean indicating whether messages at the warning level should be displayed')

    errorActive = property(lambda l: _essentia.errorLevel(), lambda l, active: _essentia.setErrorLevel(active),
                           doc = 'A boolean indicating whether messages at the error level should be displayed')

    @staticmethod
    def debug(level, s):
        _essentia.log_debug(level, s)

    @staticmethod
    def info(s):
        _essentia.log_info(s)

    @staticmethod
    def warning(s):
        _essentia.log_warning(s)

    @staticmethod
    def error(s):
        _essentia.log_error(s)

# Main logger to be used by the submodules
log = EssentiaLogger()

# FIXME: remove the use of INFO in favor of essentia.log.info
INFO = log.info

# we wrap this here so that we can do the decorator trick in all_tests.py
# FIXME: what decorator trick? is this comment still valid?
def run(gen):
    from essentia.streaming import VectorInput
    # catch this here as the actual type has not been determined yet so trying
    # run it here and now would result in an invalid pointer dereference...
    if isinstance(gen, VectorInput) and not list(gen.connections.values())[0]:
        raise EssentiaError('VectorInput is not connected to anything...')
    return _essentia.run(gen)

log.debug(EPython, 'Successfully imported essentia python module (log fully available and synchronized with the C++ one)')



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

#!/usr/bin/python

import sys
import logging

class Progress:
    def __init__(self, total, format = '%d%% done...'):
        if total < 0:
            raise IndexError('total needs to be >= 0')
        self.total = total
        self.current = 0
        self.format = '\r' + format

        # small hack: if total = 0, then we're already done, so pretend we
        # actually did something and just print 100%
        if total == 0:
            self.total = 1
            self.current = 1

    def percent(self, n):
        return n*100 / self.total

    def verbose(self):
        return logging.getLogger().isEnabledFor(logging.INFO)

    def update(self, current):
        shouldUpdate = (self.percent(current) > self.percent(self.current))
        self.current = current
        if shouldUpdate:
            self.updateDisplay()

    def updateDisplay(self):
        if self.verbose():
            print (self.format % self.percent(self.current),
                   sys.stdout.flush())

    def finish(self):
        self.update(self.total)
        # also account for the final end of line
        if self.verbose():
            print()



#!/usr/bin/env python

# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program.  If not, see http://www.gnu.org/licenses/


import argparse
from os import symlink, remove
from os.path import join, dirname, abspath
from shutil import copytree, rmtree
from subprocess import call


if __name__ == "__main__":
    parser = argparse.ArgumentParser('Sets up Tensorflow for linking against it.'
        'This can be done in two ways:'
        ' - python: By symlinking to an existing Tensorflow Python package. '
        '     This is the recommended way if Tensorflow is avaialble from Python.'
        ' - libtensorflow: By downloading and installing the C API. This mode does not '
        '     allow simultaneous use of Essentia and Tensorflow from Python due to symbol name conflicts.')

    parser.add_argument('--mode', '-m', default='python',
                        choices=['python', 'libtensorflow'])
    parser.add_argument('--platform', '-p', default='linux',
                        choices=['linux', 'macos', 'windows'])
    parser.add_argument('--context', '-c', default='/usr/local/')
    parser.add_argument('--version', '-v', default='1.14.0')
    parser.add_argument('--with_gpu', action='store_true')

    args = parser.parse_args()

    context = args.context

    if args.mode == 'python':
        print('looking for the tensorflow python package...')
        try:
            import tensorflow
        except ImportError as error:
            print(error.__class__.__name__ + ": " + error.message)
            raise(Exception('tensorflow is not available from this interpreter.\n'
                            'Suggestion: `pip install tensorflow`'))

        tf_dir = dirname(tensorflow.__file__)
        version = tensorflow.__version__

        version_list = version.split('.')
        major_version = int(version_list[0])
        minor_version = int(version_list[1])

        # From Tensorflow 1.15. libraries are stored in `tensorflow_core`
        if major_version >= 1:
            if minor_version >= 15:
                tf_dir = tf_dir + '_core'

        print('found tensorflow in "{}"'.format(tf_dir))
        print('tensorflow version: {}'.format(version))

        file_dir = dirname(abspath(__file__))

        # create symbolic links fo the libraries
        print('creating symbolic links...')

        src = join(tf_dir, 'libtensorflow_framework.so.1')
        tgt = join(context, 'lib', 'libtensorflow_framework.so')
        call(['ln', '-sf', src, tgt])
        call(['ln', '-sf', 'libtensorflow_framework.so', join(context, 'lib', 'libtensorflow_framework.so.1')])

        src = join(tf_dir, 'python', '_pywrap_tensorflow_internal.so')
        tgt = join(context, 'lib', 'libpywrap_tensorflow_internal.so')
        call(['ln', '-sf', src, tgt])
        call(['ln', '-sf', 'libpywrap_tensorflow_internal.so', join(context, 'lib', '_pywrap_tensorflow_internal.so')])

        libs = ('-lpywrap_tensorflow_internal '
                '-ltensorflow_framework')

        # copy headers to the context dir
        include_dir = join(context, 'include', 'tensorflow', 'c')
        rmtree(include_dir, ignore_errors=True)
        copytree(join(dirname(__file__), 'c'), include_dir)

    elif args.mode == 'libtensorflow':
        # WARNING. With `--mode libtensorflow`, the following problem is known
        # to arise when importing Essentia and Tensorflow in Python at the same time.
        #   In [1]: import tensorflow
        #   In [2]: import essentia
        #
        #   ImportError: /usr/local/lib/libtensorflow.so.1: undefined symbol:
        #  _ZN6google8protobuf5Arena18CreateMaybeMessageIN10tensorflow16OptimizerOptionsEIEEEPT_PS1_DpOT0_

        # The recommended solution is to link Essentia against the Tensorflow shared
        # libraries shipped with the Tensorflow wheel with `--mode python`

        print('downloading libtensorflow...')
        with_gpu = args.with_gpu
        platform = args.platform
        version = args.version

        hardware = 'gpu' if with_gpu else 'cpu'

        tarball = ('libtensorflow-{}-{}-x86_64-{}.tar.gz'.format(hardware, platform, version))
        url = ('https://storage.googleapis.com/tensorflow/libtensorflow/{}'.format(tarball))

        # download the tarball
        call(['wget', url])

        # copy it to the given context
        print('extracting...')
        call(['tar', '-C', context, '-xzf', tarball])
        remove(tarball)

        libs = '-ltensorflow'

    else:
        raise(Exception('Not valid operation mode chosen.'))

    # create the pkg-config file

    print('preparing pkg-config file...')
    includes = '-I' + join(context, 'include/tensorflow')
    lib_dirs = '-L' + join(context, 'lib')

    pkg_config = ('Name: tensorflow\n'
                  'Description: machine learning lib -- development files\n'
                  'Version: {}\n'
                  'Libs: {} {}\n'
                  'Cflags: {}\n').format(version, lib_dirs, libs, includes)

    with open(join(context, 'lib', 'pkgconfig', 'tensorflow.pc'), 'w') as f:
        f.write(pkg_config)

    # sometimes the dynamic linker has to be reconfigured
    print('reconfiguring the linker...')
    call(['ldconfig'])

    print('done!')



# Copyright (C) 2006-2020  Music Technology Group - Universitat Pompeu Fabra
#
# This file is part of Essentia
#
# Essentia is free software: you can redistribute it and/or modify it under
# the terms of the GNU Affero General Public License as published by the Free
# Software Foundation (FSF), either version 3 of the License, or (at your
# option) any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the Affero GNU General Public License
# version 3 along with this program. If not, see http://www.gnu.org/licenses/

import os
import pylab
import math


def descriptorPlot(name, xData, yData, pool, options):

    pylab.figure()
    pylab.title(name)
    pylab.plot(xData, yData)
    duration = int(math.floor(pool.value('metadata.duration_processed')))
    maxValue = max(yData)
    minValue = min(yData)

    # plotting segments lines
    if options['segmentation']['doSegmentation']:
        segments = pool.value('segmentation.timestamps')
        for segment in segments:
            pylab.plot([segment[0], segment[0]], [minValue, maxValue], 'r-')
            pylab.plot([segment[1], segment[1]], [minValue, maxValue], 'r-')

    pylab.axis([-2, duration + 2, minValue, maxValue])
    if not os.path.exists('plots'):
        os.mkdir('plots')
    figureName = 'plots/' + name + '.png'
    print('Plotting ' + name + '...')
    pylab.savefig(figureName)

    return figureName


def descriptorPlotHTML(namespace, name, audio, pool, options):

    try:
        # plot name
        descName = namespace + '.' + name
        # plot x data
        scopes = pool.value(namespace + '.' + name + '.' + 'scope')
        descScopes = []
        for scope in scopes:
            descScopes.append(scope[0])
        # plot y data
        descValues = pool.value(namespace + '.' + name)
        # plotting
        try:
            figureName = descriptorPlot(descName, descScopes, descValues, pool, options)
            htmlCode = '<img src ="' + figureName  + '"/>'
        except RuntimeError:
            # special case: descriptors with more than one dimension (mfcc, barkbands, etc...)
            htmlCode = ''
            for i in range(len(descValues[0])):
                descSubName = descName + '.' + str(i)
                descSubValues = []
                for value in descValues:
                    descSubValues.append(value[i])
                figureName = descriptorPlot(descSubName, descScopes, descSubValues, pool, options)
                htmlCode += '<img src ="' + figureName  + '"/>'

    except KeyError:
        htmlCode = ''
        print("WARNING: the descriptor", descName, "doesn't exist")

    return htmlCode


def namespace_comp(ns1, ns2):
    if ns1 == 'special': return -1
    if ns2 == 'special': return 1
    return cmp(ns1, ns2)


def compute(inputFilename, audio, pool, options):

    htmlCode = '<p align="center"><b><font size=6>' + inputFilename + '</font></b></p>'
    html = False
    descriptors = options['plotsList']
    namespaces = [ ns for ns in descriptors ]
    namespaces.sort(namespace_comp)
    htmlFile = open(inputFilename + '.html', 'w')
    htmlCode = '<p align="center"><b><font size=6>' + inputFilename + '</font></b></p>'
    html = False

    # plot signal
    descName = "signal"
    descValues = audio
    descScopes = []
    for time in range(len(audio)):
        descScopes.append(time / options['sampleRate'])
    figureName = descriptorPlot(descName, descScopes, descValues, pool, options)
    htmlCode += '<img src ="' + figureName  + '"/>'

    # plot descriptors, one by one
    for namespace in namespaces:
        names = [ n for n in descriptors[namespace] ]
        names.sort()
        for name in names:
            htmlCode += descriptorPlotHTML(namespace, name, audio, pool, options)

    # write HTML file
    htmlFile = open(inputFilename + '.html', 'w')
    htmlFile.write(htmlCode)
    htmlFile.close()




#!/usr/bin/env python

from Numeric import *
from FFT import *

def make_random(len):
    import random
    res=[]
    for i in range(int(len)):
        r=random.uniform(-1,1)
        i=random.uniform(-1,1)
        res.append( complex(r,i) )
    return res

def slowfilter(sig,h):
    translen = len(h)-1
    return convolve(sig,h)[translen:-translen]

def nextpow2(x):
    return 2 ** math.ceil(math.log(x)/math.log(2))

def fastfilter(sig,h,nfft=None):
    if nfft is None:
        nfft = int( nextpow2( 2*len(h) ) )
    H = fft( h , nfft )
    scraplen = len(h)-1
    keeplen = nfft-scraplen
    res=[]
    isdone = 0
    lastidx = nfft
    idx0 = 0
    while not isdone:
        idx1 = idx0 + nfft
        if idx1 >= len(sig):
            idx1 = len(sig)
            lastidx = idx1-idx0
            if lastidx <= scraplen:
                break
            isdone = 1
        Fss = fft(sig[idx0:idx1],nfft)
        fm = Fss * H
        m = inverse_fft(fm)
        res.append( m[scraplen:lastidx] )
        idx0 += keeplen
    return concatenate( res )

def main():
    import sys
    from getopt import getopt
    opts,args = getopt(sys.argv[1:],'rn:l:')
    opts=dict(opts)

    siglen = int(opts.get('-l',1e4 ) )
    hlen =50 
 
    nfft = int(opts.get('-n',128) )
    usereal = opts.has_key('-r')

    print 'nfft=%d'%nfft
    # make a signal
    sig = make_random( siglen )
    # make an impulse response
    h = make_random( hlen )
    #h=[1]*2+[0]*3
    if usereal:
        sig=[c.real for c in sig]
        h=[c.real for c in h]

    # perform MAC filtering
    yslow = slowfilter(sig,h)
    #print '<YSLOW>',yslow,'</YSLOW>'
    #yfast = fastfilter(sig,h,nfft)
    yfast = utilfastfilter(sig,h,nfft,usereal)
    #print yfast
    print 'len(yslow)=%d'%len(yslow)
    print 'len(yfast)=%d'%len(yfast)
    diff = yslow-yfast
    snr = 10*log10( abs( vdot(yslow,yslow) / vdot(diff,diff) ) )
    print 'snr=%s' % snr
    if snr < 10.0:
        print 'h=',h
        print 'sig=',sig[:5],'...'
        print 'yslow=',yslow[:5],'...'
        print 'yfast=',yfast[:5],'...'

def utilfastfilter(sig,h,nfft,usereal):
    import compfft
    import os
    open( 'sig.dat','w').write( compfft.dopack(sig,'f',not usereal) )
    open( 'h.dat','w').write( compfft.dopack(h,'f',not usereal) )
    if usereal: 
        util = './fastconvr' 
    else:
        util = './fastconv'
    cmd = 'time %s -n %d -i sig.dat -h h.dat -o out.dat' % (util, nfft)
    print cmd
    ec  = os.system(cmd)
    print 'exited->',ec
    return compfft.dounpack(open('out.dat').read(),'f',not usereal)

if __name__ == "__main__":
    main()
